"""
Salt Module to perform CRUD operations for NSX-T's IP Address Pools
"""
import logging

from saltext.vmware.utils import common
from saltext.vmware.utils import nsxt_request

log = logging.getLogger(__name__)

__virtual_name__ = "nsxt_ip_pools"

IP_POOLS_BASE_URL = "https://{}/api/v1/pools/ip-pools"


def __virtual__():
    return __virtual_name__


def get(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    cursor=None,
    included_fields=None,
    page_size=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Lists all IP Address pools present in the NSX-T Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_ip_pools.get hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    page_size
        (Optional) Maximum number of results to return in this page

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order

    """

    log.info("Fetching IP Address Pools")
    url = IP_POOLS_BASE_URL.format(hostname)

    params = common._filter_kwargs(
        allowed_kwargs=["cursor", "included_fields", "page_size", "sort_ascending", "sort_by"],
        default_dict=None,
        cursor=cursor,
        included_fields=included_fields,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )

    return nsxt_request.call_api(
        method="get",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_display_name(
    hostname, username, password, display_name, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Gets IP Address pool present in the NSX-T Manager with given name.

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_ip_pools.get_by_display_name hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        The name of IP Address pool to fetch

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """

    log.info("Finding IP Address Pool with display name: %s", display_name)

    ip_pools = common._read_paginated(
        func=get,
        display_name=display_name,
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in ip_pools:
        return ip_pools

    return {"results": ip_pools}


def create(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    display_name=None,
    description=None,
    tags=None,
    subnets=None,
    ip_release_delay=None,
):
    """
    Creates an IP Address pool with given specifications

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_ip_pools.create hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    display_name
        The name using which IP Address Pool will be created. If not provided then pool id will be used as
        display name

    description
        (Optional) description for the IP Address Pool

    tags
        (Optional) Opaque identifiers meaningful to the API user. Maximum 30 tags can be associated:

        .. code::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    subnets
        (Optional) The collection of one or more subnet objects in a pool.
        Subnets can be IPv4 or IPv6 and they should not overlap. The maximum number will not exceed 5 subnets.

        .. code::

            subnets='[
                {
                    "cidr": "cidr_value",
                    "gateway_ip": "gateway_ip_value",
                    "dns_nameservers": [
                        "dns_nameserver1",
                        "dns_nameserver2"
                    ],
                    "allocation_ranges": [
                        {
                            "start": "IP-Address-Range-start",
                            "end": "IP-Address-Range-end"
                        }
                    ]
                }
            ]'

    ip_release_delay
        (Optional) Delay in milliseconds, while releasing allocated IP address from IP pool (Default is 2 mins - configured on NSX device).
    """

    log.info("Creating IP Address Pool")
    url = IP_POOLS_BASE_URL.format(hostname)

    req_data = common._filter_kwargs(
        allowed_kwargs=["display_name", "description", "subnets", "tags", "ip_release_delay"],
        default_dict=None,
        display_name=display_name,
        description=description,
        tags=tags,
        subnets=subnets,
        ip_release_delay=ip_release_delay,
    )

    return nsxt_request.call_api(
        method="post",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=req_data,
    )


def update(
    ip_pool_id,
    display_name,
    revision,
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    description=None,
    tags=None,
    subnets=None,
    ip_release_delay=None,
):
    """
    Updates an IP Address pool of display name with given specifications, All the fields for which no value is
    provided will be set to null

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_ip_pools.update hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    ip_pool_id
        Id of the existing IP Address pool

    display_name
        Existing IP Pool display name. This is a non updatable field

    description
        (Optional) description for the IP Address Pool

    tags
        (Optional) Opaque identifiers meaningful to the API user. Maximum 30 tags can be associated:

        .. code::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    subnets
        (Optional) The collection of one or more subnet objects in a pool.
        Subnets can be IPv4 or IPv6 and they should not overlap. The maximum number will not exceed 5 subnets.

        .. code::

            subnets='[
                {
                    "cidr": "cidr_value",
                    "gateway_ip": "gateway_ip_value",
                    "dns_nameservers": [
                        "dns_nameserver1",
                        "dns_nameserver2"
                    ],
                    "allocation_ranges": [
                        {
                            "start": "IP-Address-Range-start",
                            "end": "IP-Address-Range-end"
                        }
                    ]
                }
            ]'

    ip_release_delay
        (Optional) Delay in milliseconds, while releasing allocated IP address from IP pool (Default is 2 mins).

    revision
        Revision number of IP Pool to update
    """

    log.info("Updating IP Address Pool %s", display_name)
    url = IP_POOLS_BASE_URL.format(hostname) + "/{}".format(ip_pool_id)

    req_data = common._filter_kwargs(
        allowed_kwargs=["description", "subnets", "tags", "ip_release_delay"],
        default_dict={
            "id": ip_pool_id,
            "_revision": revision,
            "display_name": display_name,
        },
        description=description,
        tags=tags,
        subnets=subnets,
        ip_release_delay=ip_release_delay,
    )

    return nsxt_request.call_api(
        method="put",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=req_data,
    )


def delete(
    ip_pool_id, hostname, username, password, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Deletes an IP Address pool with given id

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_ip_pools.delete hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    ip_pool_id
        Existing IP Pool id

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """

    log.info("Deleting IP Address Pool %s", ip_pool_id)

    url = IP_POOLS_BASE_URL.format(hostname) + "/{}".format(ip_pool_id)

    response = nsxt_request.call_api(
        method="delete",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    return response or "IP Pool deleted successfully"



"""
Salt Module to perform CRUD operations for NSX-T transport node profiles
"""
import logging

from saltext.vmware.utils import common
from saltext.vmware.utils import nsxt_request

log = logging.getLogger(__name__)

__virtual_name__ = "nsxt_transport_node_profiles"


def __virtual__():
    return __virtual_name__


create_params_for_transport_profiles = [
    "transport_zone_endpoints",
    "description",
    "ignore_overridden_hosts",
]


def get(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    cursor=None,
    included_fields=None,
    page_size=None,
    sort_ascending=None,
    sort_by=None,
):
    """

    Lists NSX-T transport node profiles

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_node_profiles.get hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    page_size
        (Optional) Maximum number of results to return in this page

    sort_ascending
        (Optional) Boolean. Specifies sorting order

    sort_by
        (Optional) Field by which records are sorted

    For more information, see `VMware API docs for NSX-T <https://code.vmware.com/apis/1163/nsx-t>`_

    """
    log.info("Fetching NSX-T transport node profiles")
    url = f"https://{hostname}/api/v1/transport-node-profiles"
    params = common._filter_kwargs(
        allowed_kwargs=["cursor", "included_fields", "page_size", "sort_ascending", "sort_by"],
        default_dict=None,
        cursor=cursor,
        included_fields=included_fields,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )
    return nsxt_request.call_api(
        method="get",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_display_name(
    hostname,
    username,
    password,
    display_name,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """

    Gets nsxt transport node profiles present in the NSX-T Manager with given display_name.

    CLI Example:
    .. code-block:: bash

        salt vm_minion nsxt_transport_node_profiles.get_by_display_name hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        The name of the transport node profile

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against certificate common name

    """
    log.info("Finding transport node profiles with display name: %s", display_name)
    transport_node_profiles = common._read_paginated(
        func=get,
        display_name=display_name,
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )
    if "error" in transport_node_profiles:
        return transport_node_profiles
    return {"results": transport_node_profiles}


def create(
    hostname,
    username,
    password,
    display_name,
    host_switch_spec,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    transport_zone_endpoints=None,
    description=None,
    ignore_overridden_hosts=None,
):
    """

    Creates transport node profile with the data payload provided.

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_node_profiles.create hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        The name of the transport node profile

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against certificate common name

    description
        (Optional) Description of this resource

    display_name
        (Optional) Display name for the transport node profile. Defaults to nsxt id, if not set

    host_switch_spec
        (Optional) Transport node host switch specification
        The HostSwitchSpec is the base class for standard and preconfigured host switch specifications.
        Only standard host switches are supported in the transport node profile.

    transport_zone_endpoints
        (Optional) This is deprecated. TransportZoneEndPoints should be specified per host switch at
        StandardHostSwitch through Transport Node or Transport Node Profile configuration. Array of transport zones.

    ignore_overridden_hosts
        (Optional) Boolean which determines if cluster-level configuration should be applied on overridden hosts. Default: False

        .. code-block:: yaml

            hostname: <nsxt-hostname>
            username: <nsxt-username>
            password: <nsxt-password>
            verify-ssl: <True/False>
            display_name: <display_name>
            description: <description>
            host_switch_spec:
              resource_type: StandardHostSwitchSpec
              host_switches:
                - host_switch_name: nvds1
                  host_switch_type: NVDS/VDS
                  host_switch_mode: STANDARD
                  host_switch_profile_ids:
                    -  key: UplinkHostSwitchProfile
                       value: <Respective nsxt id>
                    -  key: LldpHostSwitchProfile
                       value: <Respective nsxt id>
                    -  key: NiocProfile
                       value: <Respective nsxt id>
                  is_migrate_pnics: false
                  ip_assignment_spec:
                    resource_type: AssignedByDhcp
                  transport_zone_endpoints:
                    -  transport_zone_id: <Respective nsxt id>

    """
    log.info("Creating nsxt transport node profile")
    url = f"https://{hostname}/api/v1/transport-node-profiles"
    req_data = common._filter_kwargs(
        allowed_kwargs=create_params_for_transport_profiles,
        default_dict={},
        transport_zone_endpoints=transport_zone_endpoints,
        description=description,
        ignore_overridden_hosts=ignore_overridden_hosts,
    )
    req_data["resource_type"] = "TransportNodeProfile"
    req_data["display_name"] = display_name
    req_data["host_switch_spec"] = host_switch_spec
    req_data["host_switch_spec"]["resource_type"] = "StandardHostSwitchSpec"
    return nsxt_request.call_api(
        method="post",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=req_data,
    )


def update(
    hostname,
    username,
    password,
    display_name,
    host_switch_spec,
    transport_node_profile_id,
    revision,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    transport_zone_endpoints=None,
    description=None,
    ignore_overridden_hosts=None,
):
    """

    Updates transport node profile with the data payload provided.

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_node_profiles.update hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        (Optional) The name of the transport node profile

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against certificate common name

    transport_node_profile_id
        Unique id provided by NSX-T for transport node profile

    revision
        _revision property of the transport node profile provided by NSX-T

    description
        (Optional) Description of this resource

    display_name
        (Optional) Display name for the transport node profile. Defaults to nsxt id, if not set

    host_switch_spec
        (Optional) Transport node host switch specification
        The HostSwitchSpec is the base class for standard and preconfigured host switch specifications.
        Only standard host switches are supported in the transport node profile.

    transport_zone_endpoints
        (Optional) This is deprecated. TransportZoneEndPoints should be specified per host switch at
        StandardHostSwitch through Transport Node or Transport Node Profile configuration. Array of transport zones.

    ignore_overridden_hosts
        (Optional) Boolean which determines if cluster-level configuration should be applied on overridden hosts. Default: False

        .. code-block:: yaml

            hostname: <nsxt-hostname>
            username: <nsxt-username>
            password: <nsxt-password>
            verify-ssl: <True/False>
            display_name: <display_name>
            description: <description>
            host_switch_spec:
              resource_type: StandardHostSwitchSpec
              host_switches:
                - host_switch_name: nvds1
                  host_switch_type: NVDS/VDS
                  host_switch_mode: STANDARD
                  host_switch_profile_ids:
                    -  key: UplinkHostSwitchProfile
                       value: <Respective nsxt id>
                    -  key: LldpHostSwitchProfile
                       value: <Respective nsxt id>
                    -  key: NiocProfile
                       value: <Respective nsxt id>
                  is_migrate_pnics: false
                  ip_assignment_spec:
                    resource_type: AssignedByDhcp
                  transport_zone_endpoints:
                    -  transport_zone_id: <Respective nsxt id>

    """
    log.info("Updating nsxt transport node profile")
    url = f"https://{hostname}/api/v1/transport-node-profiles/{transport_node_profile_id}"
    req_data = common._filter_kwargs(
        allowed_kwargs=create_params_for_transport_profiles,
        default_dict={},
        transport_zone_endpoints=transport_zone_endpoints,
        description=description,
        ignore_overridden_hosts=ignore_overridden_hosts,
    )
    req_data["resource_type"] = "TransportNodeProfile"
    req_data["display_name"] = display_name
    req_data["host_switch_spec"] = host_switch_spec
    req_data["host_switch_spec"]["resource_type"] = "StandardHostSwitchSpec"
    req_data["_revision"] = revision
    return nsxt_request.call_api(
        method="put",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=req_data,
    )


def delete(
    hostname,
    username,
    password,
    transport_node_profile_id,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """

    Deletes transport node profile

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_node_profiles.delete hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    transport_node_profile_id
        Existing transport node profile id

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against certificate common name

    """
    log.info("Deleting transport node profile with id %s", transport_node_profile_id)
    url = f"https://{hostname}/api/v1/transport-node-profiles/{transport_node_profile_id}"
    result = nsxt_request.call_api(
        method="delete",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )
    return result or {"message": "Deleted transport node profile successfully"}



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging


log = logging.getLogger(__name__)

try:
    import pyVmomi

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_datacenter"
__proxyenabled__ = ["vmware_datacenter"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def present(name):
    """
    Create a datacenter

    .. code-block:: yaml

        Create Datacenter:
          vmware_datacenter.present:
            - name: dc1
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    dcs = __salt__["vmware_datacenter.list"]()
    if name in dcs:
        ret["comment"] = "Datacenter {} is already present. No changes made.".format(name)
        ret["result"] = True
    elif __opts__["test"]:
        ret["comment"] = "Datacenter {} will be created.".format(name)
        ret["result"] = None
    else:
        dc = __salt__["vmware_datacenter.create"](name)
        if isinstance(dc, dict) and dc.get(name) is not False:
            ret["comment"] = "Datacenter - {} created.".format(name)
            ret["changes"] = dc
            ret["result"] = True

        else:
            ret["comment"] = dc["reason"]
            ret["result"] = False
    return ret


def absent(name):
    """
    Delete a datacenter.

    .. code-block:: yaml

        Delete Datacenter:
          vmware_datacenter.absent:
            - name: dc1
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    dcs = __salt__["vmware_datacenter.list"]()
    if name not in dcs:
        ret["comment"] = "Datacenter {} does not exist. No changes made.".format(name)
        ret["result"] = True
    elif __opts__["test"]:
        ret["comment"] = "Datacenter {} will be deleted.".format(name)
        ret["result"] = None
    else:
        dc = __salt__["vmware_datacenter.delete"](name)
        if isinstance(dc, dict) and dc.get(name) is not False:
            ret["comment"] = "Datacenter - {} deleted.".format(name)
            ret["changes"] = dc
            ret["result"] = True

        else:
            ret["comment"] = dc["reason"]
            ret["result"] = False
    return ret



"""
VMC nat rules state module

Add new nat rule, update existing nat rule and delete existing nat rule from an SDDC.

Example usage :

.. code-block:: yaml

    ensure_nat_rule:
      vmc_nat_rules.present:
        - hostname: sample-nsx.vmwarevmc.com
        - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
        - authorization_host: console-stg.cloud.vmware.com
        - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
        - sddc_id: b43da080-2626-f64c-88e8-7f31d9d2c306
        - domain_id: mgw
        - nat_rule: vCenter_Inbound_Rule_2
        - verify_ssl: False
        - cert: /path/to/client/certificate
        - source_network: "203.0.113.73"
        - translated_network: "198.51.100.1"


.. warning::

    It is recommended to pass the VMC authentication details using Pillars rather than specifying as plain text in SLS
    files.


"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_state

log = logging.getLogger(__name__)


def present(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    tier1,
    nat,
    nat_rule,
    verify_ssl=True,
    cert=None,
    action=None,
    destination_network=None,
    source_network=None,
    translated_network=None,
    translated_ports=vmc_constants.VMC_NONE,
    scope=None,
    service=None,
    enabled=None,
    firewall_match=None,
    logging=None,
    description=None,
    tags=vmc_constants.VMC_NONE,
    sequence_number=None,
    display_name=None,
):
    """
    Ensure a given nat rule exists for given SDDC

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the nat rules should be added

    domain_id
        The domain_id for which the nat rules should belongs to. Possible values: mgw, cgw

    nat_rule
        Id of the nat rule to be added to SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    action
        specify type of nat rule it can have value REFLEXIVE, DNAT

            REFLEXIVE nat rule require
                source_network
                translated_network
                service should be empty
                translated_ports  should be None

            DNAT  Rule require
                service
                destination_network
                translated_network
                translated_ports
                source_network can be None or input network.

    destination_network
        Represents the destination network
            This supports single IP address or comma separated list of single IP
            addresses or CIDR. This does not support IP range or IP sets.

    scope
        Array of policy paths of labels, ProviderInterface, NetworkInterface

    service
        Represents the service on which the NAT rule will be applied

    source_network
        Represents the source network address
            This supports single IP address or comma separated list of single IP
            addresses or CIDR. This does not support IP range or IP sets.

    translated_network
        Represents the translated network address

            This supports single IP address or comma separated list of single IP
            addresses or CIDR. This does not support IP range or IP sets.

    translated_ports
        Port number or port range

            Please note, if there is service configured in this NAT rule, the translated_port
            will be realized on NSX Manager as the destination_port. If there is no sevice configured,
            the port will be ignored.
    enabled
        (Optional) Policy nat rule enabled flag

            The flag, which suggests whether the NAT rule is enabled or
            disabled. The default is True.

    firewall_match
        (Optional) Represents the firewall match flag

            It indicates how the firewall matches the address after NATing if firewall
            stage is not skipped.
            possible values: MATCH_EXTERNAL_ADDRESS, MATCH_INTERNAL_ADDRESS
            Default: "MATCH_INTERNAL_ADDRESS"

    logging
        (Optional) Policy nat rule logging flag
            default: False

    description
        (Optional) Description of of nat rule

    tags
        (Optional) Opaque identifiers meaningful to the API user. Maximum 30 tags can be associated:

        .. code-block::

            tags:
              - tag: <tag-key-1>
                scope: <tag-value-1>
              - tag: <tag-key-2>
                scope: <tag-value-2>

    sequence_number
        (Optional) Sequence number of the nat rule
            The sequence_number decides the rule_priority of a NAT rule.
            default: 0
            type: int

    display_name
        Identifier to use when displaying entity in logs or GUI. This is applicable for only update scenario.
        For create scenario, display_name would be same as rule_id.

    Example Values:

        .. code-block::

            action: REFLEXIVE
            translated_network: 203.0.113.36
            translated_ports: null
            destination_network: ''
            source_network: 198.51.100.23
            sequence_number: 0
            service: ''
            logging: false
            enabled: false
            scope:
              - /infra/labels/cgw-public
            tags:
              - tag: tag1
                scope: scope1
            description: ''
            firewall_match: MATCH_INTERNAL_ADDRESS
    """

    input_dict = {
        "action": action,
        "description": description,
        "destination_network": destination_network,
        "scope": scope,
        "service": service,
        "source_network": source_network,
        "tags": tags,
        "translated_network": translated_network,
        "translated_ports": translated_ports,
        "enabled": enabled,
        "firewall_match": firewall_match,
        "logging": logging,
        "sequence_number": sequence_number,
        "display_name": display_name,
    }

    input_dict = {k: v for k, v in input_dict.items() if v != vmc_constants.VMC_NONE}

    get_nat_rule_response = __salt__["vmc_nat_rules.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        tier1=tier1,
        nat=nat,
        nat_rule=nat_rule,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in get_nat_rule_response:
        if "could not be found" in get_nat_rule_response["error"]:
            get_nat_rule_response = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=get_nat_rule_response["error"], result=False
            )

    if __opts__.get("test"):
        log.info("present is called with test option")
        return vmc_state._create_state_response(
            name=name,
            comment="State present will {} nat rule {}".format(
                "update" if get_nat_rule_response else "create", nat_rule
            ),
        )

    if get_nat_rule_response:
        updatable_keys = input_dict.keys()
        is_update_required = vmc_state._check_for_updates(
            get_nat_rule_response, input_dict, updatable_keys, ["translated_ports", "tags"]
        )

        if is_update_required:
            updated_nat_rule = __salt__["vmc_nat_rules.update"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                tier1=tier1,
                nat=nat,
                nat_rule=nat_rule,
                verify_ssl=verify_ssl,
                cert=cert,
                action=action,
                destination_network=destination_network,
                source_network=source_network,
                translated_network=translated_network,
                translated_ports=translated_ports,
                scope=scope,
                service=service,
                enabled=enabled,
                firewall_match=firewall_match,
                logging=logging,
                description=description,
                tags=tags,
                sequence_number=sequence_number,
                display_name=display_name,
            )

            if "error" in updated_nat_rule:
                return vmc_state._create_state_response(
                    name=name, comment=updated_nat_rule["error"], result=False
                )

            updated_nat_rule = __salt__["vmc_nat_rules.get_by_id"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                tier1=tier1,
                nat=nat,
                nat_rule=nat_rule,
                verify_ssl=verify_ssl,
                cert=cert,
            )

            if "error" in updated_nat_rule:
                return vmc_state._create_state_response(
                    name=name, comment=updated_nat_rule["error"], result=False
                )

            return vmc_state._create_state_response(
                name=name,
                comment="Updated nat rule {}".format(nat_rule),
                old_state=get_nat_rule_response,
                new_state=updated_nat_rule,
                result=True,
            )
        else:
            log.info("All fields are same as existing nat rule %s", nat_rule)
            return vmc_state._create_state_response(
                name=name, comment="Nat rule exists already, no action to perform", result=True
            )
    else:
        log.info("No nat rule found with Id %s", nat_rule)
        created_nat_rule = __salt__["vmc_nat_rules.create"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            tier1=tier1,
            nat=nat,
            nat_rule=nat_rule,
            verify_ssl=verify_ssl,
            cert=cert,
            action=action,
            destination_network=destination_network,
            source_network=source_network,
            translated_network=translated_network,
            translated_ports=translated_ports,
            scope=scope,
            service=service,
            enabled=enabled,
            firewall_match=firewall_match,
            logging=logging,
            description=description,
            tags=tags,
            sequence_number=sequence_number,
        )

        if "error" in created_nat_rule:
            return vmc_state._create_state_response(
                name=name, comment=created_nat_rule["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Created nat rule {}".format(nat_rule),
            new_state=created_nat_rule,
            result=True,
        )


def absent(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    tier1,
    nat,
    nat_rule,
    verify_ssl=True,
    cert=None,
):
    """
    Ensure a given nat rule does not exist on given SDDC

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the nat rule should be deleted

    domain_id
        The domain_id for which the nat rules should belongs to. Possible values: mgw, cgw

    nat_rule
        Id of the nat rule to be deleted from SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Checking if nat rule with Id %s is present", nat_rule)
    get_nat_rule_response = __salt__["vmc_nat_rules.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        tier1=tier1,
        nat=nat,
        nat_rule=nat_rule,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in get_nat_rule_response:
        if "could not be found" in get_nat_rule_response["error"]:
            get_nat_rule_response = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=get_nat_rule_response["error"], result=False
            )

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if get_nat_rule_response:
            return vmc_state._create_state_response(
                name=name, comment="State absent will delete nat rule with Id {}".format(nat_rule)
            )
        else:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will do nothing as no nat rule found with Id {}".format(
                    nat_rule
                ),
            )

    if get_nat_rule_response:
        log.info("Security found with Id %s", nat_rule)
        deleted_nat_rule = __salt__["vmc_nat_rules.delete"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            tier1=tier1,
            nat=nat,
            nat_rule=nat_rule,
            verify_ssl=verify_ssl,
            cert=cert,
        )

        if "error" in deleted_nat_rule:
            return vmc_state._create_state_response(
                name=name, comment=deleted_nat_rule["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Deleted nat rule {}".format(nat_rule),
            old_state=get_nat_rule_response,
            result=True,
        )
    else:
        log.info("No nat rule found with Id %s", nat_rule)
        return vmc_state._create_state_response(
            name=name, comment="No nat rule found with Id {}".format(nat_rule), result=True
        )



"""
VMC Security group state module

Add new security group, update existing security group and delete existing security group from an SDDC.

Example usage :

.. code-block:: yaml

    Security_group_1:
      vmc_security_groups.present:
        - hostname: sample-nsx.vmwarevmc.com
        - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
        - authorization_host: console-stg.cloud.vmware.com
        - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
        - sddc_id: b43da080-2626-f64c-88e8-7f31d9d2c306
        - domain_id: mgw
        - verify_ssl: False
        - cert: /path/to/client/certificate

.. warning::

    It is recommended to pass the VMC authentication details using Pillars rather than specifying as plain text in SLS
    files.


"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_state

log = logging.getLogger(__name__)


def present(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    verify_ssl=True,
    cert=None,
    expression=None,
    description=None,
    tags=vmc_constants.VMC_NONE,
    display_name=None,
):
    """
    Ensure a given security group exists for given SDDC

    name
        Indicates the security group id, any unique string identifying the security group.
        Also same as the display_name by default.

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the security groups should be added

    domain_id
        The domain_id for which the security group should belong to. Possible values: mgw, cgw

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.
    expression
        Expression for security group members
            The expression list must follow below criteria:
            1. A non-empty expression list, must be of odd size. In a list, with
            indices starting from 0, all non-conjunction expressions must be at
            even indices, separated by a conjunction expression at odd
            indices.
            2. The total of ConditionExpression and NestedExpression in a list
            should not exceed 5.
            3. The total of IPAddressExpression, MACAddressExpression, external
            IDs in an ExternalIDExpression and paths in a PathExpression must not exceed
            500.
            4. Each expression must be a valid Expression. See the definition of
            the Expression type for more information.

        Its list of dicts
            Example values

            1. [{"member_type":"VirtualMachine","resource_type":"ExternalIDExpression",
            "external_ids":["52bf8bd0-95b1-2e58-5180-ccfa743da576"]}]

            2. [{"value":"Linux","member_type":"VirtualMachine","key":"OSName",
               "operator":"EQUALS","resource_type":"Condition"},
               {"resource_type":"ConjunctionOperator","conjunction_operator":"OR"},
               {"member_type":"VirtualMachine","resource_type":"ExternalIDExpression",
               "external_ids":["52bf8bd0-95b1-2e58-5180-ccfa743da576"]}]

            3. [{"ip_addresses" : ["203.0.113.1", "203.0.113.2"],
                "resource_type" : "IPAddressExpression"} ]

    description
        (Optional) Description of Security Groups

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code-block::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    display_name
        Identifier to use when displaying entity in logs or GUI. This is applicable for only update scenario.
        For create scenario, display_name would be same as security_group_id.

    Example values:

        .. code-block::

            {
                "expression": [
                  {
                    "member_type": "VirtualMachine",
                    "value": "webvm",
                    "key": "Tag",
                    "operator": "EQUALS",
                    "resource_type": "Condition"
                  }
                ],
                "description": "web group"
            }

    """
    security_group_id = name
    input_dict = {
        "expression": expression,
        "description": description,
        "tags": tags,
        "display_name": display_name,
    }

    input_dict = {k: v for k, v in input_dict.items() if v != vmc_constants.VMC_NONE}

    security_group = __salt__["vmc_security_groups.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_group_id=security_group_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in security_group:
        if "could not be found" in security_group["error"]:
            security_group = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=security_group["error"], result=False
            )

    if __opts__.get("test"):
        log.info("present is called with test option")
        return vmc_state._create_state_response(
            name=name,
            comment="Security group {} will be {}".format(
                security_group_id, "updated" if security_group else "created"
            ),
        )

    if security_group:
        updatable_keys = input_dict.keys()
        is_update_required = vmc_state._check_for_updates(
            security_group, input_dict, updatable_keys, ["tags"]
        )

        if is_update_required:
            updated_security_group = __salt__["vmc_security_groups.update"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                domain_id=domain_id,
                security_group_id=security_group_id,
                verify_ssl=verify_ssl,
                cert=cert,
                expression=expression,
                description=description,
                tags=tags,
                display_name=display_name,
            )

            if "error" in updated_security_group:
                return vmc_state._create_state_response(
                    name=name, comment=updated_security_group["error"], result=False
                )

            updated_security_group = __salt__["vmc_security_groups.get_by_id"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                domain_id=domain_id,
                security_group_id=security_group_id,
                verify_ssl=verify_ssl,
                cert=cert,
            )

            if "error" in updated_security_group:
                return vmc_state._create_state_response(
                    name=name, comment=updated_security_group["error"], result=False
                )

            return vmc_state._create_state_response(
                name=name,
                comment="Updated security group {}".format(security_group_id),
                old_state=security_group,
                new_state=updated_security_group,
                result=True,
            )
        else:
            log.info("All fields are same as existing Security group %s", security_group_id)
            return vmc_state._create_state_response(
                name=name,
                comment="Security group exists already, no action to perform",
                result=True,
            )
    else:
        log.info("No security group found with ID %s", security_group_id)
        created_security_group = __salt__["vmc_security_groups.create"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            domain_id=domain_id,
            security_group_id=security_group_id,
            verify_ssl=verify_ssl,
            cert=cert,
            expression=expression,
            description=description,
            tags=tags,
        )

        if "error" in created_security_group:
            return vmc_state._create_state_response(
                name=name, comment=created_security_group["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Created security group {}".format(security_group_id),
            new_state=created_security_group,
            result=True,
        )


def absent(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    verify_ssl=True,
    cert=None,
):
    """
    Ensure a given security group does not exist on given SDDC

    name
        Indicates the security group id, any unique string identifying the security group.

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the security group should be deleted

    domain_id
        The domain_id for which the security group should belong to. Possible values: mgw, cgw

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    security_group_id = name
    log.info("Checking if security group with ID %s is present", security_group_id)
    security_group = __salt__["vmc_security_groups.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_group_id=security_group_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in security_group:
        if "could not be found" in security_group["error"]:
            security_group = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=security_group["error"], result=False
            )

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if security_group:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will delete security group with ID {}".format(
                    security_group_id
                ),
            )
        else:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will do nothing as no security group found with ID {}".format(
                    security_group_id
                ),
            )

    if security_group:
        log.info("Security group found with ID %s", security_group_id)
        deleted_security_group = __salt__["vmc_security_groups.delete"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            domain_id=domain_id,
            security_group_id=security_group_id,
            verify_ssl=verify_ssl,
            cert=cert,
        )

        if "error" in deleted_security_group:
            return vmc_state._create_state_response(
                name=name, comment=deleted_security_group["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Deleted security group {}".format(security_group_id),
            old_state=security_group,
            result=True,
        )
    else:
        log.info("No security group found with ID %s", security_group_id)
        return vmc_state._create_state_response(
            name=name,
            comment="No security group found with ID {}".format(security_group_id),
            result=True,
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Common functions used across modules
"""
import json


def drift_report(obj1, obj2, diff_level=None):
    """
    Finds the drift between two objects/dictionaries

    obj1
        First dictionary
    obj2
        Second dictionary
    diff_level
        Defines on what tree level to show the drift.
        If not specified the changes are shown in the last/leaf level.

        Drift level None - all leaf nodes changes:
        o
        |_
        | x
        |__
        |   |__
        o   |   x
            |__
            |   |__
            o   |   x

        Drift level 0 - changes in first level of config tree:
        o
        |
        x...
        |
        x...

        Drift level 1 - changes in second level of config tree:
        o
        |__
        o   x...
        |
        |__
        o   x...

        Drift level 2 - changes in third level of config tree:
        o
        |__
        o   o
        |   |__
        |       x...
        |__
        o   o
            |__
                x...
    """
    result_diffs = []
    _drift_recurse_(obj1, obj2, result_diffs)

    result_tree = {}
    for tup in result_diffs:
        _drift_tree_(tup, result_tree)

    if diff_level is not None and isinstance(result_tree, dict):
        result_subtree = {}
        _drift_subtree_(result_tree, diff_level, result_subtree)
        return result_subtree

    return result_tree


def _drift_tree_(diffs, result_tree):
    branch = result_tree
    for diff in diffs[:-2]:
        if diff not in branch:
            branch[diff] = {}
        branch = branch[diff]
    branch[diffs[-2]] = diffs[-1]


def _drift_subtree_(result_tree, diff_level, result_subtree, level=0, new_subtree=0):
    for k in result_tree.keys():
        if isinstance(result_tree[k], dict):
            if level == diff_level:
                result_subtree[k] = {"old": {}, "new": {}}
                _drift_subtree_(
                    result_tree[k], diff_level, result_subtree[k]["old"], level + 1, new_subtree=1
                )
                _drift_subtree_(
                    result_tree[k], diff_level, result_subtree[k]["new"], level + 1, new_subtree=2
                )
            else:
                if k not in result_subtree:
                    result_subtree[k] = {}
                _drift_subtree_(
                    result_tree[k], diff_level, result_subtree[k], level + 1, new_subtree
                )
        else:
            if new_subtree == 1:
                result_subtree[k] = result_tree[k][0]
            elif new_subtree == 2:
                result_subtree[k] = result_tree[k][1]
            else:
                result_subtree[k] = {"old": result_tree[k][0], "new": result_tree[k][1]}


def _drift_recurse_(obj1, obj2, result, keys=[]):
    if isinstance(obj1, dict) and isinstance(obj2, dict):
        if not keys:
            # use symmetric difference for first level only
            for k in set(obj1).symmetric_difference(obj2):
                if k in obj1:
                    # first level element value should be dict
                    result.append(tuple(keys) + (k,) + ((obj1[k], {}),))
                else:
                    # first level element value should be dict
                    result.append(tuple(keys) + (k,) + (({}, obj2[k]),))
        else:
            # otherwise make difference only for added elements in obj2
            for k in set(obj1).symmetric_difference(obj2):
                if k in obj2:
                    # only for complex types - dict, list
                    if isinstance(obj2[k], dict):
                        result.append(tuple(keys) + (k,) + (({}, obj2[k]),))
                    elif isinstance(obj2[k], (list, tuple)):
                        result.append(tuple(keys) + (k,) + (([], obj2[k]),))
        for k in set(obj1).intersection(obj2):
            _drift_recurse_(obj1[k], obj2[k], result, keys=keys + [k])
    else:
        if isinstance(obj1, (list, tuple)) and isinstance(obj2, (list, tuple)):
            if set(obj1) != set(obj2):
                result.append(tuple(keys) + ((obj1, obj2),))
        elif obj1 != obj2:
            result.append(tuple(keys) + ((obj1, obj2),))



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

import salt.exceptions
import saltext.vmware.utils.cluster as utils_cluster
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.datacenter as utils_datacenter

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_cluster"
__proxyenabled__ = ["vmware_cluster"]
__func_alias__ = {"list_": "list"}


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def list_(service_instance=None, profile=None, datacenter=None):
    """
    Returns a dictionary containing a list of clusters for each datacenter.

    .. code-block:: bash

        salt '*' vmware_cluster.list
    """
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    try:
        datacenters = utils_datacenter.get_datacenters(
            service_instance,
            datacenter_names=[datacenter] if datacenter else None,
            get_all_datacenters=not bool(datacenter),
        )
        for datacenter in datacenters:
            clusters = utils_common.get_mors_with_properties(
                service_instance,
                vim.ClusterComputeResource,
                container_ref=datacenter,
                property_list=["name"],
            )
            for cluster in clusters:
                ret.setdefault(datacenter.name, []).append(cluster.get("name"))
    except (salt.exceptions.VMwareApiError, salt.exceptions.VMwareObjectRetrievalError) as exc:
        log.error("Error retrieving cluster/datacenter - %s", str(exc))
    return ret


def create(name, datacenter, service_instance=None, profile=None):
    """
    Creates a cluster.

    Supported proxies: esxcluster

    name
        The cluster name

    datacenter
        The datacenter name in which the cluster is to be created

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_cluster.create dc1 cluster1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    try:
        dc_ref = utils_datacenter.get_datacenter(service_instance, datacenter)
        cluster_spec = vim.cluster.ConfigSpecEx()
        utils_cluster.create_cluster(cluster_name=name, dc_ref=dc_ref, cluster_spec=cluster_spec)
    except (salt.exceptions.VMwareApiError, salt.exceptions.VMwareRuntimeError) as exc:
        return {name: False, "reason": str(exc)}
    return {name: True}


def get(cluster_name, datacenter_name, service_instance=None, profile=None):
    """
    Get the properties of a cluster.

    cluster_name
        The cluster name

    datacenter_name
        The datacenter name to which the cluster belongs

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_cluster.get cluster_name=cl1 datacenter_name=dc1
    """
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    try:
        dc_ref = utils_datacenter.get_datacenter(service_instance, datacenter_name)
        cluster_ref = utils_cluster.get_cluster(dc_ref=dc_ref, cluster=cluster_name)

        # DRS config
        ret["drs_enabled"] = cluster_ref.configurationEx.drsConfig.enabled
        if ret["drs_enabled"]:
            ret["drs"] = __salt__["vmware_cluster_drs.get"](
                cluster_name=cluster_name,
                datacenter_name=datacenter_name,
                service_instance=service_instance,
            )

        # HA config
        ret["ha_enabled"] = cluster_ref.configurationEx.dasConfig.enabled
        if ret["ha_enabled"]:
            ret["ha"] = __salt__["vmware_cluster_ha.get"](
                cluster_name=cluster_name,
                datacenter_name=datacenter_name,
                service_instance=service_instance,
            )

        # vSAN
        ret["vsan_enabled"] = cluster_ref.configurationEx.vsanConfigInfo.enabled

    except (salt.exceptions.VMwareApiError, salt.exceptions.VMwareObjectRetrievalError) as exc:
        return {cluster_name: False, "reason": str(exc)}
    return ret


def delete(name, datacenter, service_instance=None, profile=None):
    """
    Deletes a cluster.

    Supported proxies: esxcluster

    name
        The cluster name

    datacenter
        The datacenter name to which the cluster belongs

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_cluster.delete cl1 dc1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    try:
        utils_cluster.delete_cluster(service_instance, name, datacenter)
    except (salt.exceptions.VMwareApiError, salt.exceptions.VMwareObjectRetrievalError) as exc:
        return {name: False, "reason": str(exc)}
    return {name: True}



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Common functions used across modules
"""
import errno
import logging
import re
import time
from http.client import BadStatusLine

import salt.exceptions
import salt.modules.cmdmod
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils

try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

CAMELCASE_PATTERN = re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")

log = logging.getLogger(__name__)


def camel_to_snake_case(attrib):
    return CAMELCASE_PATTERN.sub(r"_\1", attrib).lower()


def get_root_folder(service_instance):
    """
    Returns the root folder of a vCenter.

    service_instance
        The Service Instance Object for which to obtain the root folder.

    """
    try:
        log.trace("Retrieving root folder")
        return service_instance.RetrieveContent().rootFolder
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)


def get_service_content(service_instance):
    """
    Returns the service content for a Service Instance.

    service_instance
        The Service Instance from which to obtain service content.
    """
    try:
        return service_instance.RetrieveServiceContent()
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            f"Not enough permissions. Required privilege: '{exc.privilegeId}'"
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)


def get_content(
    service_instance,
    obj_type,
    property_list=None,
    container_ref=None,
    traversal_spec=None,
    local_properties=False,
):
    """
    Returns the content of the specified type of object for a Service Instance.

    For more information, please see:
    http://pubs.vmware.com/vsphere-50/index.jsp?topic=%2Fcom.vmware.wssdk.pg.doc_50%2FPG_Ch5_PropertyCollector.7.6.html

    service_instance
        The Service Instance from which to obtain content.

    obj_type
        The type of content to obtain.

    property_list
        An optional list of object properties to used to return even more filtered content results.

    container_ref
        An optional reference to the managed object to search under. Can either be an object of type Folder, Datacenter,
        ComputeResource, Resource Pool or HostSystem. If not specified, default behaviour is to search under the inventory
        rootFolder.

    traversal_spec
        An optional TraversalSpec to be used instead of the standard
        ``Traverse All`` spec.

    local_properties
        Flag specifying whether the properties to be retrieved are local to the
        container. If that is the case, the traversal spec needs to be None.
    """
    # Start at the rootFolder if container starting point not specified
    if not container_ref:
        container_ref = get_root_folder(service_instance)

    # By default, the object reference used as the starting poing for the filter
    # is the container_ref passed in the function
    obj_ref = container_ref
    local_traversal_spec = False
    if not traversal_spec and not local_properties:
        local_traversal_spec = True
        # We don't have a specific traversal spec override so we are going to
        # get everything using a container view
        try:
            obj_ref = service_instance.content.viewManager.CreateContainerView(
                container_ref, [obj_type], True
            )
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.RuntimeFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareRuntimeError(exc.msg)

        # Create 'Traverse All' traversal spec to determine the path for
        # collection
        traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
            name="traverseEntities",
            path="view",
            skip=False,
            type=vim.view.ContainerView,
        )

    # Create property spec to determine properties to be retrieved
    property_spec = vmodl.query.PropertyCollector.PropertySpec(
        type=obj_type, all=True if not property_list else False, pathSet=property_list
    )

    # Create object spec to navigate content
    obj_spec = vmodl.query.PropertyCollector.ObjectSpec(
        obj=obj_ref,
        skip=True if not local_properties else False,
        selectSet=[traversal_spec] if not local_properties else None,
    )

    # Create a filter spec and specify object, property spec in it
    filter_spec = vmodl.query.PropertyCollector.FilterSpec(
        objectSet=[obj_spec],
        propSet=[property_spec],
        reportMissingObjectsInResults=False,
    )

    # Retrieve the contents
    try:
        content = service_instance.content.propertyCollector.RetrieveContents([filter_spec])
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)

    # Destroy the object view
    if local_traversal_spec:
        try:
            obj_ref.Destroy()
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.RuntimeFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareRuntimeError(exc.msg)

    return content


def get_mors_with_properties(
    service_instance,
    object_type,
    property_list=None,
    container_ref=None,
    traversal_spec=None,
    local_properties=False,
):
    """
    Returns a list containing properties and managed object references for the managed object.

    service_instance
        The Service Instance from which to obtain managed object references.

    object_type
        The type of content for which to obtain managed object references.

    property_list
        An optional list of object properties used to return even more filtered managed object reference results.

    container_ref
        An optional reference to the managed object to search under. Can either be an object of type Folder, Datacenter,
        ComputeResource, Resource Pool or HostSystem. If not specified, default behaviour is to search under the inventory
        rootFolder.

    traversal_spec
        An optional TraversalSpec to be used instead of the standard
        ``Traverse All`` spec

    local_properties
        Flag specifying whether the properties to be retrieved are local to the
        container. If that is the case, the traversal spec needs to be None.
    """
    # Get all the content
    content_args = [service_instance, object_type]
    content_kwargs = {
        "property_list": property_list,
        "container_ref": container_ref,
        "traversal_spec": traversal_spec,
        "local_properties": local_properties,
    }
    try:
        content = get_content(*content_args, **content_kwargs)
    except BadStatusLine:
        content = get_content(*content_args, **content_kwargs)
    except OSError as exc:
        if exc.errno != errno.EPIPE:
            raise
        content = get_content(*content_args, **content_kwargs)

    object_list = []
    for obj in content:
        properties = {}
        for prop in obj.propSet:
            properties[prop.name] = prop.val
        properties["object"] = obj.obj
        object_list.append(properties)
    log.trace("Retrieved %s objects", len(object_list))
    return object_list


def get_mor_by_property(
    service_instance,
    object_type,
    property_value,
    property_name="name",
    container_ref=None,
):
    """
    Returns the first managed object reference having the specified property value.

    service_instance
        The Service Instance from which to obtain managed object references.

    object_type
        The type of content for which to obtain managed object references.

    property_value
        The name of the property for which to obtain the managed object reference.

    property_name
        An object property used to return the specified object reference results. Defaults to ``name``.

    container_ref
        An optional reference to the managed object to search under. Can either be an object of type Folder, Datacenter,
        ComputeResource, Resource Pool or HostSystem. If not specified, default behaviour is to search under the inventory
        rootFolder.
    """
    # Get list of all managed object references with specified property
    object_list = get_mors_with_properties(
        service_instance,
        object_type,
        property_list=[property_name],
        container_ref=container_ref,
    )

    for obj in object_list:
        obj_id = str(obj.get("object", "")).strip("'\"")
        if obj[property_name] == property_value or property_value == obj_id:
            return obj["object"]

    return None


def list_objects(service_instance, vim_object, properties=None):
    """
    Returns a simple list of objects from a given service instance.

    service_instance
        The Service Instance for which to obtain a list of objects.

    object_type
        The type of content for which to obtain information.

    properties
        An optional list of object properties used to return reference results.
        If not provided, defaults to ``name``.
    """
    if properties is None:
        properties = ["name"]

    items = []
    item_list = get_mors_with_properties(service_instance, vim_object, properties)
    for item in item_list:
        if item.get("name"):
            items.append(item["name"])
    return items


def get_service_instance_from_managed_object(mo_ref, name="<unnamed>"):
    """
    Retrieves the service instance from a managed object.

    me_ref
        Reference to a managed object (of type vim.ManagedEntity).

    name
        Name of managed object. This field is optional.
    """
    if not name:
        name = mo_ref.name
    log.trace("[%s] Retrieving service instance from managed object", name)
    si = vim.ServiceInstance("ServiceInstance")
    si._stub = mo_ref._stub
    return si


def get_properties_of_managed_object(mo_ref, properties):
    """
    Returns specific properties of a managed object, retrieved in an
    optimally.

    mo_ref
        The managed object reference.

    properties
        List of properties of the managed object to retrieve.
    """
    service_instance = get_service_instance_from_managed_object(mo_ref)
    log.trace("Retrieving name of %s", type(mo_ref).__name__)
    try:
        items = get_mors_with_properties(
            service_instance,
            type(mo_ref),
            container_ref=mo_ref,
            property_list=["name"],
            local_properties=True,
        )
        mo_name = items[0]["name"]
    except vmodl.query.InvalidProperty:
        mo_name = "<unnamed>"
    log.trace(
        "Retrieving properties '%s' of %s '%s'",
        properties,
        type(mo_ref).__name__,
        mo_name,
    )
    items = get_mors_with_properties(
        service_instance,
        type(mo_ref),
        container_ref=mo_ref,
        property_list=properties,
        local_properties=True,
    )
    if not items:
        raise salt.exceptions.VMwareApiError(
            "Properties of managed object '{}' weren't " "retrieved".format(mo_name)
        )
    return items[0]


def get_managed_object_name(mo_ref):
    """
    Returns the name of a managed object.
    If the name wasn't found, it returns None.

    mo_ref
        The managed object reference.
    """
    props = get_properties_of_managed_object(mo_ref, ["name"])
    return props.get("name")


def get_resource_pools(
    service_instance,
    resource_pool_names,
    datacenter_name=None,
    get_all_resource_pools=False,
):
    """
    Retrieves resource pool objects

    service_instance
        The service instance object to query the vCenter

    resource_pool_names
        Resource pool names

    datacenter_name
        Name of the datacenter where the resource pool is available

    get_all_resource_pools
        Boolean

    return
        Resourcepool managed object reference
    """

    properties = ["name"]
    if not resource_pool_names:
        resource_pool_names = []
    if datacenter_name:
        import saltext.vmware.utils.datacenter as utils_datacenter

        container_ref = utils_datacenter.get_datacenter(service_instance, datacenter_name)
    else:
        container_ref = get_root_folder(service_instance)

    resource_pools = get_mors_with_properties(
        service_instance,
        vim.ResourcePool,
        container_ref=container_ref,
        property_list=properties,
    )

    selected_pools = []
    for pool in resource_pools:
        if get_all_resource_pools or (pool["name"] in resource_pool_names):
            selected_pools.append(pool["object"])
    if not selected_pools:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "The resource pools with properties "
            "names={} get_all={} could not be found".format(selected_pools, get_all_resource_pools)
        )

    return selected_pools


def _filter_kwargs(allowed_kwargs, default_dict=None, **kwargs):
    result = default_dict or {}
    for field in allowed_kwargs:
        val = kwargs.get(field)
        if val is not None:
            result[field] = val
    return result


def _read_paginated(func, display_name, **kwargs):
    results = []
    paginated = {"cursor": None}
    while "cursor" in paginated:
        paginated = func(**kwargs)
        if "error" in paginated:
            return paginated
        results.extend(
            result for result in paginated["results"] if result.get("display_name") == display_name
        )
        kwargs["cursor"] = paginated.get("cursor")
    return results


def wait_for_task(task, instance_name, task_type, sleep_seconds=1, log_level="debug"):
    """
    Waits for a task to be completed.

    task
        The task to wait for.

    instance_name
        The name of the ESXi host, vCenter Server, or Virtual Machine that
        the task is being run on.

    task_type
        The type of task being performed. Useful information for debugging purposes.

    sleep_seconds
        The number of seconds to wait before querying the task again.
        Defaults to ``1`` second.

    log_level
        The level at which to log task information. Default is ``debug``,
        but ``info`` is also supported.
    """
    time_counter = 0
    start_time = time.time()
    log.trace("task = %s, task_type = %s", task, task.__class__.__name__)
    try:
        task_info = task.info
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.FileNotFound as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareFileNotFoundError(exc.msg)
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    while task_info.state == "running" or task_info.state == "queued":
        if time_counter % sleep_seconds == 0:
            msg = "[ {} ] Waiting for {} task to finish [{} s]".format(
                instance_name, task_type, time_counter
            )
            if log_level == "info":
                log.info(msg)
            else:
                log.debug(msg)
        time.sleep(1.0 - ((time.time() - start_time) % 1.0))
        time_counter += 1
        try:
            task_info = task.info
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.FileNotFound as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareFileNotFoundError(exc.msg)
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.RuntimeFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareRuntimeError(exc.msg)
    if task_info.state == "success":
        msg = "[ {} ] Successfully completed {} task in {} seconds".format(
            instance_name, task_type, time_counter
        )
        if log_level == "info":
            log.info(msg)
        else:
            log.debug(msg)
        # task is in a successful state
        return task_info.result
    else:
        # task is in an error state
        try:
            raise task_info.error
        except vim.fault.NoPermission as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(
                "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
            )
        except vim.fault.FileNotFound as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareFileNotFoundError(exc.msg)
        except vim.fault.VimFault as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareApiError(exc.msg)
        except vmodl.fault.SystemError as exc:
            log.exception(exc)
            raise salt.exceptions.VMwareSystemError(exc.msg)
        except vmodl.fault.InvalidArgument as exc:
            log.exception(exc)
            exc_message = exc.msg
            if exc.faultMessage:
                exc_message = "{} ({})".format(exc_message, exc.faultMessage[0].message)
            raise salt.exceptions.VMwareApiError(exc_message)


def get_parent_type(node, parent_type):
    """
    Return a parent of specified type from a node.

    node
        Object reference to start at.

    parent_type
        The vim type of the parent you are searching for.
    """
    if isinstance(node, parent_type):
        return node
    try:
        node = node.parent
    except AttributeError:
        return None
    return get_parent_type(node, parent_type)


def get_path(node, service_instance, path=""):
    """
    Return a path to root from a node.

    node
        Object reference to start at.

    service_instance
        The Service Instance from which to obtain managed object references.

    path
        Path to node, recursively passed.
    """
    if node == service_instance.content.rootFolder:
        return path
    try:
        path = "/" + node.name + path
        node = node.parent
    except AttributeError:
        return path
    return get_path(node, service_instance, path)


def get_datacenters(service_instance, datacenter_names=None, get_all_datacenters=False):
    """
    Returns all datacenters in a vCenter.

    service_instance
        The Service Instance Object from which to obtain cluster.

    datacenter_names
        List of datacenter names to filter by. Default value is None.

    get_all_datacenters
        Flag specifying whether to retrieve all datacenters.
        Default value is None.
    """
    items = [
        i["object"]
        for i in get_mors_with_properties(service_instance, vim.Datacenter, property_list=["name"])
        if get_all_datacenters or (datacenter_names and i["name"] in datacenter_names)
    ]
    return items


def get_datacenter(service_instance, datacenter_name):
    """
    Returns a vim.Datacenter managed object.

    service_instance
        The Service Instance Object from which to obtain datacenter.

    datacenter_name
        The datacenter name
    """
    items = get_datacenters(service_instance, datacenter_names=[datacenter_name])
    if not items:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "Datacenter '{}' was not found".format(datacenter_name)
        )
    return items[0]


def list_datacenters(service_instance):
    """
    Returns a list of datacenters associated with a given service instance.

    service_instance
        The Service Instance Object from which to obtain datacenters.
    """
    return list_objects(service_instance, vim.Datacenter)


def create_datacenter(service_instance, datacenter_name):
    """
    Creates a datacenter.

    .. versionadded:: 2017.7.0

    service_instance
        The Service Instance Object

    datacenter_name
        The datacenter name
    """
    root_folder = get_root_folder(service_instance)
    log.trace("Creating datacenter '%s'", datacenter_name)
    try:
        dc_obj = root_folder.CreateDatacenter(datacenter_name)
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    return dc_obj


def delete_datacenter(service_instance, datacenter_name):
    """
    Deletes a datacenter.

    service_instance
        The Service Instance Object

    datacenter_name
        The datacenter name
    """
    root_folder = get_root_folder(service_instance)
    log.trace("Deleting datacenter '%s'", datacenter_name)
    try:
        dc_obj = get_datacenter(service_instance, datacenter_name)
        task = dc_obj.Destroy_Task()
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    wait_for_task(task, datacenter_name, "DeleteDatacenterTask")


def get_parent_of_type(mors, type):
    """
    Finds the first parent of a managed object that matches the type specified.

    `None` is returned if no object is found.
    """
    while True:
        if isinstance(mors, type):
            return mors
        try:
            mors = mors.parent
        except AttributeError:
            return None


def find_filtered_object(service_instance, datacenter_name=None, cluster_name=None, host_name=None):
    """
    Finds zero or one matching objects: plug in almost any combination of datacenter, cluster, and/or host name.

    If cluster_name is passed, datacenter_name must also be passed.

    At least one of the optional parameters must be set.

    The most specific object will be returned (if you pass host_name and datacenter_name, the host will be returned).

    service_instance
        The Service Instance Object from which to obtain cluster.

    datacenter_name
        (Optional) Datacenter name to filter by.

    cluster_name
        (Optional) Exact cluster name to filter by. If used, datacenter_name is required.

    host_name
        (Optional) Exact host name name to filter by.
    """
    try:
        if host_name:
            import saltext.vmware.utils.esxi as utils_esxi

            hosts = utils_esxi.get_hosts(
                service_instance,
                datacenter_name=datacenter_name,
                cluster_name=cluster_name,
                host_names=[host_name],
            )
            return hosts[0] if hosts else None
        elif cluster_name and datacenter_name:
            import saltext.vmware.utils.cluster as utils_cluster

            datacenter = get_datacenter(service_instance, datacenter_name)
            return utils_cluster.get_cluster(datacenter, cluster_name)
        elif datacenter_name:
            return get_datacenter(service_instance, datacenter_name=datacenter_name)
        else:
            raise salt.exceptions.ArgumentValueError(
                "find_filtered_object requires at least one of datacenter_name, host_name, or cluster_name with datacenter_name"
            )
    except salt.exceptions.VMwareObjectRetrievalError:
        return None


def get_license_mgrs(service_instance, license_mgr_names=None, get_all_license_mgrs=False):
    """
    Returns all license managers in a vCenter.

    service_instance
        The Service Instance Object from which to obtain cluster.

    license_mgr_names
        List of license manager names to filter by. Default value is None.

    get_all_license_mgrs
        Flag specifying whether to retrieve all license managers.
        Default value is None.
    """
    log.debug("started get all License Managers")
    items = [
        i["object"]
        for i in get_mors_with_properties(
            service_instance, vim.LicenseAssignmentManager, property_list=["name"]
        )
        if get_all_license_mgrs or (license_mgr_names and i["name"] in license_mgr_names)
    ]
    log.debug("exited get all License Managers")
    return items


def get_license_mgr(service_instance, license_mgr_name):
    """
    Returns a vim.LicenseAssignmentManager managed object.

    service_instance
        The Service Instance Object from which to obtain license manager.

    license_mgr_name
        The license manager name
    """
    log.debug(f"started get License Manager '{license_mgr_name}'")
    items = get_license_mgrs(service_instance, license_mgr_names=[license_mgr_name])
    if not items:
        raise salt.exceptions.VMwareObjectRetrievalError(
            f"license manager '{license_mgr_name}' was not found"
        )
    log.debug(f"exit License Manager '{license_mgr_name}'")
    return items[0]


def list_license_mgrs(service_instance):
    """
    Returns a list of license managers associated with a given service instance.

    service_instance
        The Service Instance Object from which to obtain license managers.
    """
    log.debug("start list of License Managers")
    return list_objects(service_instance, vim.LicenseAssignmentManager)


def deployment_resources(host_name, service_instance):
    """
    Returns the dict representation of deployment resources from given host name.

    host_name
        The name of the esxi host to obtain esxi reference.

    """
    destination_host_ref = get_mor_by_property(
        service_instance,
        vim.HostSystem,
        host_name,
    )
    datacenter_ref = get_parent_type(destination_host_ref, vim.Datacenter)
    cluster_ref = get_parent_type(destination_host_ref, vim.ClusterComputeResource)
    resource_pool = cluster_ref.resourcePool

    return {
        "destination_host": destination_host_ref,
        "datacenter": datacenter_ref,
        "cluster": cluster_ref,
        "resource_pool": resource_pool,
    }


def get_storage_system(service_instance, host_ref, hostname=None):
    """
    Returns a host's storage system
    """

    if not hostname:
        hostname = get_managed_object_name(host_ref)

    traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
        path="configManager.storageSystem", type=vim.HostSystem, skip=False
    )
    objs = get_mors_with_properties(
        service_instance,
        vim.HostStorageSystem,
        property_list=["systemFile"],
        container_ref=host_ref,
        traversal_spec=traversal_spec,
    )
    if not objs:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "Host's '{}' storage system was not retrieved".format(hostname)
        )
    log.trace("[%s] Retrieved storage system", hostname)
    return objs[0]["object"]


def _get_scsi_address_to_lun_key_map(
    service_instance, host_ref, storage_system=None, hostname=None
):
    """
    Returns a map between the scsi addresses and the keys of all luns on an ESXi
    host.
        map[<scsi_address>] = <lun key>
    service_instance
        The Service Instance Object from which to obtain the hosts
    host_ref
        The vim.HostSystem object representing the host that contains the
        requested disks.
    storage_system
        The host's storage system. Default is None.
    hostname
        Name of the host. Default is None.
    """
    if not hostname:
        hostname = get_managed_object_name(host_ref)
    if not storage_system:
        storage_system = get_storage_system(service_instance, host_ref, hostname)
    try:
        device_info = storage_system.storageDeviceInfo
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: {}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    if not device_info:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "Host's '{}' storage device info was not retrieved".format(hostname)
        )
    multipath_info = device_info.multipathInfo
    if not multipath_info:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "Host's '{}' multipath info was not retrieved".format(hostname)
        )
    if multipath_info.lun is None:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "No luns were retrieved from host '{}'".format(hostname)
        )
    lun_key_by_scsi_addr = {}
    for l in multipath_info.lun:
        # The vmware scsi_address may have multiple comma separated values
        # The first one is the actual scsi address
        lun_key_by_scsi_addr.update({p.name.split(",")[0]: l.lun for p in l.path})
    log.trace("Scsi address to lun id map on host '%s': %s", hostname, lun_key_by_scsi_addr)
    return lun_key_by_scsi_addr


def get_all_luns(host_ref, storage_system=None, hostname=None):
    """
    Returns a list of all vim.HostScsiDisk objects in a disk

    host_ref
        The vim.HostSystem object representing the host that contains the requested disks.

    storage_system
        The host's storage system. Default is None.

    hostname
        Name of the host. This argument is optional.
    """
    if not hostname:
        hostname = get_managed_object_name(host_ref)
    if not storage_system:
        si = get_service_instance_from_managed_object(host_ref, name=hostname)
        storage_system = get_storage_system(si, host_ref, hostname)
        if not storage_system:
            raise salt.exceptions.VMwareObjectRetrievalError(
                "Host's '{}' storage system was not retrieved".format(hostname)
            )
    try:
        device_info = storage_system.storageDeviceInfo
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: {}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    if not device_info:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "Host's '{}' storage device info was not retrieved".format(hostname)
        )

    scsi_luns = device_info.scsiLun
    if scsi_luns:
        log.trace(
            "Retrieved scsi luns in host '%s': %s",
            hostname,
            [l.canonicalName for l in scsi_luns],
        )
        return scsi_luns
    log.trace("Retrieved no scsi_luns in host '%s'", hostname)
    return []


def get_scsi_address_to_lun_map(host_ref, storage_system=None, hostname=None):
    """
    Returns a map of all vim.ScsiLun objects on a ESXi host keyed by their scsi address

    host_ref
        The vim.HostSystem object representing the host that contains the requested disks.

    storage_system
        The host's storage system. Default is None.

    hostname
        Name of the host. This argument is optional.
    """
    if not hostname:
        hostname = get_managed_object_name(host_ref)
    si = get_service_instance_from_managed_object(host_ref, name=hostname)
    if not storage_system:
        storage_system = get_storage_system(si, host_ref, hostname)
    lun_ids_to_scsi_addr_map = _get_scsi_address_to_lun_key_map(
        si, host_ref, storage_system, hostname
    )
    luns_to_key_map = {d.key: d for d in get_all_luns(host_ref, storage_system, hostname)}
    return {
        scsi_addr: luns_to_key_map[lun_key]
        for scsi_addr, lun_key in lun_ids_to_scsi_addr_map.items()
    }


def get_disks(host_ref, disk_ids=None, scsi_addresses=None, get_all_disks=False):
    """
    Returns a list of vim.HostScsiDisk objects representing disks
    in a ESXi host, filtered by their cannonical names and scsi_addresses

    host_ref
        The vim.HostSystem object representing the host that contains the
        requested disks.

    disk_ids
        The list of canonical names of the disks to be retrieved. Default value
        is None

    scsi_addresses
        The list of scsi addresses of the disks to be retrieved. Default value
        is None

    get_all_disks
        Specifies whether to retrieve all disks in the host.
        Default value is False.
    """
    hostname = get_managed_object_name(host_ref)
    if get_all_disks:
        log.trace("Retrieving all disks in host '%s'", hostname)
    else:
        log.trace(
            "Retrieving disks in host '%s': ids = (%s); scsi addresses = (%s)",
            hostname,
            disk_ids,
            scsi_addresses,
        )
        if not (disk_ids or scsi_addresses):
            return []
    si = get_service_instance_from_managed_object(host_ref, name=hostname)
    storage_system = get_storage_system(si, host_ref, hostname)
    disk_keys = []
    if scsi_addresses:
        # convert the scsi addresses to disk keys
        lun_key_by_scsi_addr = _get_scsi_address_to_lun_key_map(
            si, host_ref, storage_system, hostname
        )
        disk_keys = [
            key for scsi_addr, key in lun_key_by_scsi_addr.items() if scsi_addr in scsi_addresses
        ]
        log.trace("disk_keys based on scsi_addresses = %s", disk_keys)

    scsi_luns = get_all_luns(host_ref, storage_system)
    scsi_disks = [
        disk
        for disk in scsi_luns
        if isinstance(disk, vim.HostScsiDisk)
        and (
            get_all_disks
            or
            # Filter by canonical name
            (disk_ids and (disk.canonicalName in disk_ids))
            or
            # Filter by disk keys from scsi addresses
            (disk.key in disk_keys)
        )
    ]
    log.trace(
        "Retrieved disks in host '%s': %s",
        hostname,
        [d.canonicalName for d in scsi_disks],
    )
    return scsi_disks


def get_diskgroups(host_ref, cache_disk_ids=None, get_all_disk_groups=False):
    """
    Returns a list of vim.VsanHostDiskMapping objects representing disks
    in a ESXi host, filtered by their cannonical names.

    host_ref
        The vim.HostSystem object representing the host that contains the
        requested disks.

    cache_disk_ids
        The list of cannonical names of the cache disks to be retrieved. The
        canonical name of the cache disk is enough to identify the disk group
        because it is guaranteed to have one and only one cache disk.
        Default is None.

    get_all_disk_groups
        Specifies whether to retrieve all disks groups in the host.
        Default value is False.
    """
    hostname = get_managed_object_name(host_ref)
    if get_all_disk_groups:
        log.trace("Retrieving all disk groups on host '%s'", hostname)
    else:
        log.trace(
            "Retrieving disk groups from host '%s', with cache disk ids : (%s)",
            hostname,
            cache_disk_ids,
        )
        if not cache_disk_ids:
            return []
    try:
        vsan_host_config = host_ref.config.vsanHostConfig
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: {}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    if not vsan_host_config:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "No host config found on host '{}'".format(hostname)
        )
    vsan_storage_info = vsan_host_config.storageInfo
    if not vsan_storage_info:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "No vsan storage info found on host '{}'".format(hostname)
        )
    vsan_disk_mappings = vsan_storage_info.diskMapping
    if not vsan_disk_mappings:
        return []
    disk_groups = [
        diskmap
        for diskmap in vsan_disk_mappings
        if (get_all_disk_groups or (diskmap.ssd.canonicalName in cache_disk_ids))
    ]
    log.trace(
        "Retrieved disk groups on host '%s', with cache disk ids : %s",
        hostname,
        [disk.ssd.canonicalName for disk in disk_groups],
    )
    return disk_groups


def get_date_time_mgr(host_reference):
    """
    Helper function that returns a dateTimeManager object
    """
    return host_reference.configManager.dateTimeSystem


def get_inventory(service_instance):
    """
    .. versionadded:: 23.4.4.0rc1

    Return the inventory of a Service Instance Object.

    service_instance
        The Service Instance Object for which to obtain inventory.
    """
    return service_instance.RetrieveContent()


def get_hardware_grains(service_instance):
    """
    .. versionadded:: 23.4.4.0rc1

    Return hardware info for standard minion grains if the service_instance is a HostAgent type

    service_instance
        The service instance object to get hardware info for
    """
    hw_grain_data = {}
    if get_inventory(service_instance).about.apiType == "HostAgent":
        view = service_instance.content.viewManager.CreateContainerView(
            service_instance.RetrieveContent().rootFolder, [vim.HostSystem], True
        )
        if view and view.view:
            hw_grain_data["manufacturer"] = view.view[0].hardware.systemInfo.vendor
            hw_grain_data["productname"] = view.view[0].hardware.systemInfo.model

            for _data in view.view[0].hardware.systemInfo.otherIdentifyingInfo:
                if _data.identifierType.key == "ServiceTag":
                    hw_grain_data["serialnumber"] = _data.identifierValue

            hw_grain_data["osfullname"] = view.view[0].summary.config.product.fullName
            hw_grain_data["osmanufacturer"] = view.view[0].summary.config.product.vendor
            hw_grain_data["osrelease"] = view.view[0].summary.config.product.version
            hw_grain_data["osbuild"] = view.view[0].summary.config.product.build
            hw_grain_data["os_family"] = view.view[0].summary.config.product.name
            hw_grain_data["os"] = view.view[0].summary.config.product.name
            hw_grain_data["mem_total"] = view.view[0].hardware.memorySize / 1024 / 1024
            hw_grain_data["biosversion"] = view.view[0].hardware.biosInfo.biosVersion
            hw_grain_data["biosreleasedate"] = (
                view.view[0].hardware.biosInfo.releaseDate.date().strftime("%m/%d/%Y")
            )
            hw_grain_data["cpu_model"] = view.view[0].hardware.cpuPkg[0].description
            hw_grain_data["kernel"] = view.view[0].summary.config.product.productLineId
            hw_grain_data["num_cpu_sockets"] = view.view[0].hardware.cpuInfo.numCpuPackages
            hw_grain_data["num_cpu_cores"] = view.view[0].hardware.cpuInfo.numCpuCores
            hw_grain_data["num_cpus"] = (
                hw_grain_data["num_cpu_sockets"] * hw_grain_data["num_cpu_cores"]
            )
            hw_grain_data["ip_interfaces"] = {}
            hw_grain_data["ip4_interfaces"] = {}
            hw_grain_data["ip6_interfaces"] = {}
            hw_grain_data["hwaddr_interfaces"] = {}
            for _vnic in view.view[0].configManager.networkSystem.networkConfig.vnic:
                hw_grain_data["ip_interfaces"][_vnic.device] = []
                hw_grain_data["ip4_interfaces"][_vnic.device] = []
                hw_grain_data["ip6_interfaces"][_vnic.device] = []

                hw_grain_data["ip_interfaces"][_vnic.device].append(_vnic.spec.ip.ipAddress)
                hw_grain_data["ip4_interfaces"][_vnic.device].append(_vnic.spec.ip.ipAddress)
                if _vnic.spec.ip.ipV6Config:
                    hw_grain_data["ip6_interfaces"][_vnic.device].append(
                        _vnic.spec.ip.ipV6Config.ipV6Address
                    )
                hw_grain_data["hwaddr_interfaces"][_vnic.device] = _vnic.spec.mac
            hw_grain_data["host"] = view.view[0].configManager.networkSystem.dnsConfig.hostName
            hw_grain_data["domain"] = view.view[0].configManager.networkSystem.dnsConfig.domainName
            hw_grain_data["fqdn"] = "{}{}{}".format(
                view.view[0].configManager.networkSystem.dnsConfig.hostName,
                ("." if view.view[0].configManager.networkSystem.dnsConfig.domainName else ""),
                view.view[0].configManager.networkSystem.dnsConfig.domainName,
            )

            for _pnic in view.view[0].configManager.networkSystem.networkInfo.pnic:
                hw_grain_data["hwaddr_interfaces"][_pnic.device] = _pnic.mac

            hw_grain_data["timezone"] = view.view[
                0
            ].configManager.dateTimeSystem.dateTimeInfo.timeZone.name
        view = None
    return hw_grain_data



"""
State module for NSX-T transport node profiles
"""
import logging

import salt.utils.dictdiffer
from saltext.vmware.utils import common
from saltext.vmware.utils import nsxt_request

log = logging.getLogger(__name__)

try:
    from saltext.vmware.modules import nsxt_transport_node_profiles

    HAS_NSXT_TRANSPORT_NODE_PROFILES = True
except ImportError:
    HAS_NSXT_TRANSPORT_NODE_PROFILES = False


def __virtual__():
    if not HAS_NSXT_TRANSPORT_NODE_PROFILES:
        return False, "'nsxt_transport_node_profiles' binary not found on system"
    return "nsxt_transport_node_profiles"


def _needs_update(existing_profile, new_profile_params):
    existing_profile_host_switch_spec = existing_profile.get("host_switch_spec")
    update_inputs_host_switch_spec = new_profile_params.get("host_switch_spec")
    if not existing_profile_host_switch_spec and update_inputs_host_switch_spec:
        return True
    if (
        existing_profile_host_switch_spec
        and update_inputs_host_switch_spec
        and salt.utils.dictdiffer.deep_diff(
            existing_profile_host_switch_spec, update_inputs_host_switch_spec
        )
    ):
        return True

    optional_params = ("ignore_overridden_hosts", "description", "transport_zone_endpoints")
    for param in optional_params:
        existing_profile_param_value = existing_profile.get(param)
        update_inputs_param_value = new_profile_params.get(param)
        if not existing_profile_param_value and update_inputs_param_value:
            return True
        if (
            existing_profile_param_value
            and update_inputs_param_value
            and existing_profile_param_value != new_profile_params.get(param)
        ):
            return True
    return False


def _get_by_display_name_result_from_module(
    hostname, username, password, verify_ssl, cert, cert_common_name, module_name, display_name
):
    module_to_call = module_name + ".get_by_display_name"
    result = __salt__[module_to_call](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        display_name=display_name,
    )
    if "error" in result:
        return result
    if len(result["results"]) > 1:
        return {
            "error": "Multiple results for {} with display_name {}".format(
                module_name, display_name
            )
        }
    if not result["results"]:
        return {"error": "No results for {} with display_name {}".format(module_name, display_name)}
    return result["results"][0]


def _get_by_display_name(
    hostname, username, password, endpoint, object_type, display_name, **kwargs
):
    url = "https://{management_host}/api/v1/{endpoint}".format(
        management_host=hostname, endpoint=endpoint
    )

    object_list = common._read_paginated(
        func=_get,
        display_name=display_name,
        url=url,
        username=username,
        password=password,
        **kwargs
    )

    if "error" in object_list:
        return object_list

    if len(object_list) > 1:
        return {
            "error": "Multiple results for {} with display_name {}".format(
                object_type, display_name
            )
        }
    if not object_list:
        return {"error": "No results for {} with display_name {}".format(object_type, display_name)}

    return object_list[0]


def _get(url, username, password, cursor=None, cert_common_name=None, verify_ssl=True, cert=None):
    params = dict()
    if cursor:
        params["cursor"] = cursor

    return nsxt_request.call_api(
        method="get",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def _update_params_with_id(
    hostname, username, password, cert, cert_common_name, verify_ssl, transport_node_profile_params
):
    for host_switch in transport_node_profile_params["host_switch_spec"]["host_switches"]:
        host_switch_type = host_switch.get("host_switch_type")
        if host_switch_type == "VDS":
            host_switch_name = host_switch.get("host_switch_name")
            if host_switch_name:
                result = _get_by_display_name(
                    hostname,
                    username,
                    password,
                    endpoint="fabric/virtual-switches",
                    object_type=host_switch_type,
                    display_name=host_switch_name,
                    cert=cert,
                    verify_ssl=verify_ssl,
                    cert_common_name=cert_common_name,
                )
                if "error" in result:
                    return result
                host_switch["host_switch_id"] = result["id"]
        host_switch_profiles = host_switch.pop("host_switch_profiles", None)
        if host_switch_profiles is not None:
            host_switch_profile_ids = []
            for host_switch_profile in host_switch_profiles:
                profile_obj = {}
                result = _get_by_display_name(
                    hostname,
                    username,
                    password,
                    endpoint="host-switch-profiles?include_system_owned=true",
                    object_type="host-switch-profiles",
                    display_name=host_switch_profile["name"],
                    cert=cert,
                    verify_ssl=verify_ssl,
                    cert_common_name=cert_common_name,
                )
                if "error" in result:
                    return result
                profile_obj["key"] = host_switch_profile["type"]
                profile_obj["value"] = result["id"]
                host_switch_profile_ids.append(profile_obj)
            host_switch["host_switch_profile_ids"] = host_switch_profile_ids

        ip_assignment_spec = host_switch.get("ip_assignment_spec")
        if ip_assignment_spec:
            ip_pool_name = ip_assignment_spec.pop("ip_pool_name", None)
            if ip_pool_name:
                result = _get_by_display_name_result_from_module(
                    hostname=hostname,
                    username=username,
                    password=password,
                    verify_ssl=verify_ssl,
                    cert=cert,
                    cert_common_name=cert_common_name,
                    module_name="nsxt_ip_pools",
                    display_name=ip_pool_name,
                )
                if "error" in result:
                    return result
                host_switch["ip_assignment_spec"]["ip_pool_id"] = result["id"]

        transport_zone_endpoints = host_switch.get("transport_zone_endpoints")
        if transport_zone_endpoints:
            for transport_zone_endpoint in transport_zone_endpoints:
                transport_zone_name = transport_zone_endpoint.pop("transport_zone_name", None)
                if transport_zone_name:
                    result = _get_by_display_name(
                        hostname,
                        username,
                        password,
                        endpoint="transport-zones",
                        object_type="transport zone",
                        display_name=transport_zone_name,
                        cert=cert,
                        verify_ssl=verify_ssl,
                        cert_common_name=cert_common_name,
                    )
                    if "error" in result:
                        return result
                    transport_zone_endpoint["transport_zone_id"] = result["id"]

        vmk_install_migrations = host_switch.get("vmk_install_migration")
        if vmk_install_migrations:
            for vmk_install_migration in vmk_install_migrations:
                destination_network_name = vmk_install_migration.pop(
                    "destination_network_name", None
                )
                if destination_network_name:
                    result = _get_by_display_name(
                        hostname,
                        username,
                        password,
                        endpoint="logical-switches",
                        object_type="destination network/logical switches",
                        display_name=destination_network_name,
                        cert=cert,
                        verify_ssl=verify_ssl,
                        cert_common_name=cert_common_name,
                    )
                    if "error" in result:
                        return result
                    vmk_install_migration["destination_network"] = result["id"]

    transport_zone_endpoints = transport_node_profile_params.get("transport_zone_endpoints")
    if transport_zone_endpoints:
        for transport_zone_endpoint in transport_zone_endpoints:
            transport_zone_name = transport_zone_endpoint.pop("transport_zone_name", None)
            if transport_zone_name:
                result = _get_by_display_name(
                    hostname,
                    username,
                    password,
                    endpoint="transport-zones",
                    object_type="transport zone",
                    display_name=transport_zone_name,
                    cert=cert,
                    verify_ssl=verify_ssl,
                    cert_common_name=cert_common_name,
                )
                if "error" in result:
                    return result
                transport_zone_endpoint["transport_zone_id"] = result["id"]
    else:
        transport_node_profile_params["transport_zone_endpoints"] = []
    return transport_node_profile_params


def present(
    name,
    hostname,
    username,
    password,
    display_name,
    host_switch_spec,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    transport_zone_endpoints=None,
    ignore_overridden_hosts=None,
    description=None,
):
    """

    Creates or Updates(if present with same display_name) transport node profiles
    Fails if multiple transport node profiles are found with same display_name

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_node_profiles.create hostname=nsxt-manager.local username=admin ...

    .. code-block:: yaml

        create_transport_node_profile:
          nsxt_transport_node_profiles.present:
            - name: Create uplink profile
              hostname: {{ pillar['nsxt_manager_hostname'] }}
              username: {{ pillar['nsxt_manager_username'] }}
              password: {{ pillar['nsxt_manager_password'] }}
              cert: {{ pillar['nsxt_manager_certificate'] }}
              verify_ssl: <False/True>
              display_name: <uplink profile name>
              description: <uplink profile description>
              host_switch_spec:
                resource_type: "StandardHostSwitchSpec"
                host_switches:
                  - host_switch_name: "hostswitch1"
                    host_switch_mode: "<STANDARD/ENS/ENS_INTERRUPT>"
                    host_switch_type: "<NVDS/VDS>"
                    host_switch_profiles:
                      - name: "<host_switch_profile_name>"
                        type: "<UplinkHostSwitchProfile/LldpHostSwitchProfile/NiocProfile>"
                    pnics:
                      - device_name: "vmnic1"
                        uplink_name: "uplink-1"
                    ip_assignment_spec:
                      resource_type: "StaticIpPoolSpec"
                      ip_pool_name: "IPPool-IPV4-1"
                    transport_zone_endpoints:
                      - transport_zone_name: "TZ1"
                    vmk_install_migration:
                      - device_name: vmk0
                        destination_network_name: "ls_vmk_Mgmt"

    name
        Name of the operation

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        The name of the transport node profile

    host_switch_spec
        Transport node host switch specification. The HostSwitchSpec is the base class for standard and preconfigured
        host switch specifications. Only standard host switches are supported in the transport node profile.

        resource_type
            Must be set to the value StandardHostSwitchSpec

        host_switches
            Array of transport Node host switches.

            host_switch_name
                Host switch name. This name will be used to reference a host switch. Default is "nsxDefaultHostSwitch"

            host_switch_mode
                Operational mode of a HostSwitch. STANDARD, ENS, ENS_INTERRUPT

            host_switch_id
                The host switch id. This ID will be used to reference a host switch.

            host_switch_profiles
                Array of host switch profiles. Each entry should have name of the profile and type of the profile

                .. code::

                    host_switch_profiles:
                      -  name: <name of the profile>
                         type: <type of the profile>

            ip_assignment_spec
                Specification for IPs to be used with host switch virtual tunnel endpoints

                resource_type
                    Resource type for IP assignment

                ip_pool_name
                    Name of the IP pool

            pnics
                Array of Physical NICs connected to the host switch

                device_name
                    device name or key

                uplink_name
                    Uplink name for this Pnic

            transport_zone_endpoints
                Array of transport zones

                transport_zone_name
                    Name of the transport zone

            vmk_install_migration
                Array of VmknicNetwork. The vmknic and logical switch mappings

                destination_network
                    The network id to which the ESX vmk interface will be migrated.

                device_name
                    ESX vmk interface name

    transport_zone_endpoints
        (Optional) This is deprecated. TransportZoneEndPoints should be specified per host switch at
        StandardHostSwitch through Transport Node or Transport Node Profile configuration. Array of transport zones.

    ignore_overridden_hosts
        (Optional) Boolean which determines if cluster-level configuration should be applied on overridden hosts. Default: False

    description
        (Optional) Description

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    transport_node_profiles_result = __salt__["nsxt_transport_node_profiles.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        display_name=display_name,
    )
    if "error" in transport_node_profiles_result:
        ret["result"] = False
        ret["comment"] = "Failed to get transport node profiles from NSX-T Manager : {}".format(
            transport_node_profiles_result["error"]
        )
        return ret

    result_count = len(transport_node_profiles_result.get("results"))
    if result_count > 1:
        ret["result"] = False
        ret["comment"] = (
            "Found multiple results(result_count={count}) for transport node profiles with "
            "display_name {display_name}".format(count=result_count, display_name=display_name)
        )
        return ret

    if __opts__["test"]:
        if result_count == 0:
            ret["result"] = None
            ret["comment"] = "Transport node profile will be created in NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "Transport node profile would be updated in NSX-T Manager"
        return ret

    existing_transport_node_profile = (
        None if result_count == 0 else transport_node_profiles_result.get("results")[0]
    )

    # transport_node_profile_params dict will be populated appropriately and passed to execution module for create/updates
    transport_node_profile_params = common._filter_kwargs(
        allowed_kwargs=("description", "transport_zone_endpoints", "ignore_overridden_hosts"),
        default_dict={"display_name": display_name, "host_switch_spec": host_switch_spec},
        description=description,
        transport_zone_endpoints=transport_zone_endpoints,
        ignore_overridden_hosts=ignore_overridden_hosts,
    )

    current_transport_node_profile = _update_params_with_id(
        hostname,
        username,
        password,
        cert,
        cert_common_name,
        verify_ssl,
        transport_node_profile_params,
    )
    if "error" in current_transport_node_profile:
        ret["result"] = False
        ret["comment"] = "Failed while creating payload for transport node profile: {}".format(
            current_transport_node_profile["error"]
        )
        return ret
    current_transport_node_profile["resource_type"] = "TransportNodeProfile"
    current_transport_node_profile["ignore_overridden_hosts"] = ignore_overridden_hosts or False

    if existing_transport_node_profile is None:
        # create transport node profile
        log.info(
            "Creating new transport node profile as no results were found in NSX-T with display_name %s",
            display_name,
        )
        create_result = __salt__["nsxt_transport_node_profiles.create"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            display_name=display_name,
            host_switch_spec=current_transport_node_profile["host_switch_spec"],
            transport_zone_endpoints=transport_zone_endpoints,
            ignore_overridden_hosts=ignore_overridden_hosts,
            description=description,
        )

        if create_result and "error" in create_result:
            ret["result"] = False
            ret["comment"] = "Failed to create transport node profile due to error: {}".format(
                create_result["error"]
            )
            return ret

        ret["comment"] = "Created transport node profile {display_name}".format(
            display_name=display_name
        )
        ret["changes"]["new"] = create_result
        return ret
    else:
        update_required = _needs_update(
            existing_transport_node_profile, current_transport_node_profile
        )
        if update_required:
            log.info("Updating transport node profile with display_name %s", display_name)
            update_result = __salt__["nsxt_transport_node_profiles.update"](
                hostname=hostname,
                username=username,
                password=password,
                verify_ssl=verify_ssl,
                cert=cert,
                cert_common_name=cert_common_name,
                display_name=display_name,
                host_switch_spec=current_transport_node_profile["host_switch_spec"],
                transport_node_profile_id=existing_transport_node_profile["id"],
                revision=existing_transport_node_profile["_revision"],
                transport_zone_endpoints=transport_zone_endpoints,
                ignore_overridden_hosts=ignore_overridden_hosts,
                description=description,
            )
            if update_result and "error" in update_result:
                ret["result"] = False
                ret["comment"] = "Failure while updating transport node profile: {}".format(
                    update_result["error"]
                )
                return ret
            ret["comment"] = "Updated transport node profile {display_name} successfully".format(
                display_name=display_name
            )
            ret["changes"]["old"] = existing_transport_node_profile
            ret["changes"]["new"] = update_result
            return ret
        else:
            log.info(
                "Update is not required. Transport node profile with same params already exists"
            )
            ret[
                "comment"
            ] = "Transport node profile already exists with similar params. No action to perform"
            return ret


def absent(
    name,
    hostname,
    username,
    password,
    display_name,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """

    Deletes transport node profile with provided display_name in NSX-T Manager, if present.
    Fails if multiple transport node profiles are found with same display_name

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_node_profiles.absent hostname=nsxt-manager.local username=admin ...

    -- code-block :: yaml

        delete_transport_node_profile:
          nsxt_transport_node_profiles.absent:
            - name: <Name of the operation>
              hostname: {{ pillar['nsxt_manager_hostname'] }}
              username: {{ pillar['nsxt_manager_username'] }}
              password: {{ pillar['nsxt_manager_password'] }}
              cert: {{ pillar['nsxt_manager_certificate'] }}
              display_name: <display_name of the transport node profile>
              verify_ssl: <False/True>


    name
        Name of the operation to perform

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        Display name of the transport node profile to delete

    cert
        Path to the SSL certificate file to connect to NSX-T manager

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
         verification. If the client certificate common name and hostname do not match (in case of self-signed
         certificates), specify the certificate common name as part of this parameter. This value is then used to
         compare against certificate common name.

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    transport_node_profiles_result = __salt__["nsxt_transport_node_profiles.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        display_name=display_name,
    )
    if "error" in transport_node_profiles_result:
        ret["result"] = False
        ret["comment"] = "Failed to get transport node profiles from NSX-T Manager : {}".format(
            transport_node_profiles_result["error"]
        )
        return ret

    result_count = len(transport_node_profiles_result.get("results"))
    if result_count > 1:
        ret["result"] = False
        ret["comment"] = (
            "Found multiple results(result_count={count}) for transport node profiles with "
            "display_name {display_name}".format(count=result_count, display_name=display_name)
        )
        return ret

    if __opts__["test"]:
        if result_count == 0:
            ret["result"] = None
            ret[
                "comment"
            ] = "No transport node profile with display_name: {} found in NSX-T Manager".format(
                display_name
            )
        else:
            ret["result"] = None
            ret["comment"] = "Transport node profile with display_name: {} will be deleted".format(
                display_name
            )
        return ret

    if result_count == 0:
        ret[
            "comment"
        ] = "No transport node profile with display_name: {} found in NSX-T Manager".format(
            display_name
        )
        return ret
    else:
        transport_node_profile_to_delete = transport_node_profiles_result.get("results")[0]
        delete_result = __salt__["nsxt_transport_node_profiles.delete"](
            hostname=hostname,
            username=username,
            password=password,
            transport_node_profile_id=transport_node_profile_to_delete["id"],
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
        )
        if "error" in delete_result:
            ret["result"] = False
            ret["comment"] = "Failed to delete transport node profile : {}".format(
                delete_result["error"]
            )
            return ret
        else:
            ret[
                "comment"
            ] = "Transport node profile with display_name: {} successfully deleted".format(
                display_name
            )
            ret["changes"]["old"] = transport_node_profile_to_delete
            ret["changes"]["new"] = {}
            return ret



"""
State module for NSX-T segment
"""
import logging

log = logging.getLogger(__name__)


try:
    from saltext.nsxt.modules import nsxt_policy_segment

    HAS_POLICY_SEGMENT = True
except ImportError:
    HAS_POLICY_SEGMENT = False


def __virtual__():
    if not HAS_POLICY_SEGMENT:
        return False, "'nsxt_policy_segment' binary not found on system"
    return "nsxt_policy_segment"


def present(
    name,
    hostname,
    username,
    password,
    display_name,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    state=None,
    tags=None,
    description=None,
    address_bindings=None,
    admin_state=None,
    advanced_config=None,
    bridge_profiles=None,
    connectivity_path=None,
    dhcp_config_path=None,
    extra_configs=None,
    l2_extension=None,
    tier0_display_name=None,
    tier0_id=None,
    tier1_display_name=None,
    tier1_id=None,
    domain_name=None,
    transport_zone_display_name=None,
    transport_zone_id=None,
    enforcementpoint_id=None,
    site_id=None,
    vlan_ids=None,
    subnets=None,
    segment_ports=None,
):
    """
      Creates or Updates(if present with same display_name) segment and its sub-resources with the given
      specifications.

      Note: To delete any subresource of segment state parameter as absent

      CLI Example:

      .. code-block:: bash

          salt vm_minion nsxt_policy_segment.present hostname=nsxt-manager.local username=admin ...

      .. code-block:: yaml

    nsxt_policy_segment.present:

      - name: Create segment
        hostname: <hostname>
        username: <username>
        password: <password>
        cert: <certificate>
        verify_ssl: <False/True>
        display_name: test-seg-4
        state: present
        domain_name: dn1
        transport_zone_display_name: "1-transportzone-730"
        replication_mode: "SOURCE"
        advanced_config:

          - address_pool_display_name: small-2-pool
            connectivity: "OFF"
            hybrid: False
            local_egress: True

        admin_state: UP
        connectivity_path: "/infra/tier-1s/d082bc25-a9b2-4d13-afe5-d3cecad4b854"
        subnets:

          - gateway_address: "40.1.1.1/16"

        segment_ports:
          - display_name: test-sp-1
            state: present
            tags:

              - scope: "scope-1"
                tag: "tag-2"

            extra_configs:
              - config_pair:
                  key: key
                  value: value
            ignored_address_bindings:
              - ip_address: "203.0.113.122"
          - display_name: test-sp-2
            state: present
          - display_name: test-sp-3
            state: present

    hostname
        The host name of NSX-T manager
    username
        Username to connect to NSX-T manager
    password
        Password to connect to NSX-T manager
    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.
    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.
    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against
    display_name:
        description: Display name.If resource ID is not specified, display_name will be used as ID.
        type: str
    state:
        choices:
        - present
        - absent
        description: State can be either 'present' or 'absent'.'present' is used to create or update resource.'absent' is used to delete resource.
    tags:
        description: Opaque identifiers meaningful to the API user.
        required: False
    description:
        description: Segment description.
    address_bindings:
        description: Address bindings for the Segment
        required: False
        type: list
        elements: dict
        suboptions:

            ip_address:
                description: IP Address for port binding
                type: str
            mac_address:
                description: Mac address for port binding
                type: str
            vlan_id:
                description: VLAN ID for port binding
                type: int
    admin_state:
        description: Represents Desired state of the Segment
        required: False
        type: str
        choices:
        - UP
        - DOWN
        default: UP
    advanced_config:
        description: Advanced configuration for Segment.
        required: False
        type: dict
        suboptions:

            address_pool_id:
                description: IP address pool ID
                type: str
            address_pool_name:
                description: IP address pool display name
                type: str
            connectivity:
                description: Connectivity configuration to manually connect
                             (ON) or disconnect (OFF) a logical entity from
                             network topology. Only valid for Tier1 Segment

                type: str
            hybrid:
                description:
                    - Flag to identify a hybrid logical switch
                    - When set to true, all the ports created on this segment
                      will behave in a hybrid fashion. The hybrid port
                      indicates to NSX that the VM intends to operate in
                      underlay mode, but retains the ability to forward egress
                      traffic to the NSX overlay network. This property is only
                      applicable for segment created with transport zone type
                      OVERLAY_STANDARD. This property cannot be modified after
                      segment is created.

                type: bool
            local_egress:
                description:
                    - Flag to enable local egress
                    - This property is used to enable proximity routing with
                      local egress. When set to true, logical router interface
                      (downlink) connecting Segment to Tier0/Tier1 gateway is
                      configured with prefix-length 32.

                type: bool
            local_egress_routing_policies:
                description: An ordered list of routing policies to forward
                             traffic to the next hop.

                type: list
                elements: dict
                suboptions:

                    nexthop_address:
                        required: true
                        description: Next hop address for proximity routing
                        type: str
                    prefix_list_paths:
                        required: true
                        description:

                            - Policy path to prefix lists, max 1 element
                            - The destination address of traffic matching a
                              prefix-list is forwarded to the nexthop_address.
                              Traffic matching a prefix list with Action DENY
                              will be dropped. Individual prefix-lists
                              specified could have different actions.

                        type: list
                        elements: str
            multicast:
                description:
                    - Enable multicast on the downlink
                    - Enable multicast for a segment. Only applicable for
                      segments connected to Tier0 gateway.

                type: bool
            uplink_teaming_policy_name:
                description:
                    - Uplink Teaming Policy Name
                    - The name of the switching uplink teaming policy for the
                      Segment. This name corresponds to one of the switching
                      uplink teaming policy names listed in TransportZone
                      associated with the Segment. When this property is
                      not specified, the segment will not have a teaming policy
                      associated with it and the host switch's default teaming
                      policy will be used by MP.

                type: str
    bridge_profiles:
        description: Bridge Profile Configuration
        required: False
        type: list
        elements: dict
        suboptions:

            bridge_profile_path:
                description:
                    - Policy path to L2 Bridge profile
                    - Same bridge profile can be configured on different
                      segments. Each bridge profile on a segment must unique.

                type: str
            uplink_teaming_policy_name:
                description:
                    - Uplink Teaming Policy Name
                    - The name of the switching uplink teaming policy for the
                      bridge endpoint. This name corresponds to one of the
                      switching uplink teaming policy names listed in the
                      transport zone. When this property is not specified, the
                      teaming policy is assigned by MP.

                type: str
            vlan_ids:
                description: VLAN specification for bridge endpoint. Either
                             VLAN ID or VLAN ranges can be specified. Not both.

                type: str
            vlan_transport_zone_path:
                description:
                    - Policy path to VLAN Transport Zone
                    - VLAN transport zone should belong to the enforcment-point
                      as the transport zone specified in the segment.

                type: str
    connectivity_path:
        description: Policy path to the connecting Tier-0 or Tier-1. Valid only
                     for segments created under Infra

        required: False
        type: str
    dhcp_config_path:
        description:
            - Policy path to DHCP configuration
            - Policy path to DHCP server or relay configuration to use for all
              IPv4 & IPv6 subnets configured on this segment.

        required: False
        type: str
    extra_configs:
        description:
            - Extra configs on Segment
            - This property could be used for vendor specific configuration in
              key value string pairs, the setting in extra_configs will be
              automatically inheritted by segment ports in the Segment.

        type: list
        required: False
        elements: dict
        suboptions:

            config_pair:
                description: Key value pair in string for the configuration
                type: dict
                required: true
                suboptions:

                    key:
                        description: Key
                        type: str
                        required: true
                    value:
                        description: Value
                        type: str
                        required: true
    l2_extension:
        description: Configuration for extending Segment through L2 VPN
        required: False
        type: dict
        suboptions:

            l2vpn_paths:
                description: Policy paths corresponding to the associated L2
                         VPN sessions

                type: list
                elements: str
            local_egress:
                description: Local Egress

               type: dict
               suboptions:

                   optimized_ips:
                       description: Gateway IP for Local Egress. Local egress
                                    is enabled only when this list is not empty

                       type: list
                       elements: str
            tunnel_id:
               description: Tunnel ID

               type: int

        mac_pool_id:
            description: Allocation mac pool associated with the Segment

            type: str
        metadata_proxy_paths:
            description: Metadata Proxy Configuration Paths

            type: list
            elements: str
    tier0_display_name:
        description: Same as tier_0_id. Either one can be specified.
                     If both are specified, tier_0_id takes
                     precedence.

        required: False

        type: str
    tier0_id:
        description: The Uplink of the Policy Segment.Mutually exclusive with tier_1_id.

        type: str
    tier1_display_name:
        description: Same as tier_1_id. Either one can be specified.
                     If both are specified, tier_1_id takes
                     precedence.

        type: str
    tier1_id:
        description: The Uplink of the Policy Segment.Mutually exclusive with tier_0_id but takes precedence.
        type: str
    domain_name:
        description: Domain name associated with the Policy Segment.
        type: str
    transport_zone_display_name:
        description: Same as transport_zone_id. Either one can be specified.
                     If both are specified, transport_zone_id takes
                     precedence.

        type: str
    transport_zone_id:
        description: The TZ associated with the Policy Segment.
        type: str
    enforcementpoint_id:
        description: The EnforcementPoint ID where the TZ is located.
                     Required if transport_zone_id is specified.

        default: default

        type: str
    site_id:
        description: The site ID where the EnforcementPoint is located.
                     Required if transport_zone_id is specified.

        default: default
        type: str
    vlan_ids:
        description: VLAN ids for a VLAN backed Segment.
                     Can be a VLAN id or a range of VLAN ids specified with '-'
                     in between.

        type: list
    subnets:
        description: Subnets that belong to this Policy Segment.
        type: dict
        suboptions:

            dhcp_ranges:
                description: DHCP address ranges for dynamic IP allocation.
                             DHCP address ranges are used for dynamic IP
                             allocation. Supports address range and CIDR
                             formats. First valid host address from the first
                             value is assigned to DHCP server IP address.
                             Existing values cannot be deleted or modified, but
                             additional DHCP ranges can be added.
                             Formats, e.g. 203.0.113.64/26,
                             203.0.113.2-203.0.113.50

                type: list
            gateway_address:
                description: Gateway IP address.
                             Gateway IP address in CIDR format for both IPv4
                             and IPv6.

                type: str
    segment_ports:
        type: list
        description: Add the Segment Ports to be create, updated, or deleted in this section
        element: dict
        suboptions:

            address_bindings:
                description: Static address binding used for the port.
                type: list
                elements: dict
                suboptions:
                ip_address:

                    description: IP Address for port binding.
                    type: str

                mac_address:
                    description: Mac address for port binding.
                    type: str

                vlan_id:
                    description: VLAN ID for port binding.
                    type: str

            attachment:

                description: VIF attachment.
                type: dict
                suboptions:

                    allocate_addresses:
                        description: Indicate how IP will be
                                    allocated for the port.

                        type: str
                        choices:

                            - IP_POOL
                            - MAC_POOL
                            - BOTH
                            - NONE
                    app_id:
                        description: ID used to identify/look up a
                                     child attachment behind a
                                     parent attachment.

                        type: str
                    context_id:
                        description: Parent VIF ID if type is CHILD,
                                     Transport node ID if type is
                                     INDEPENDENT.

                        type: str
                    id:
                        description: VIF UUID on NSX Manager.
                        type: str
                    traffic_tag:
                        description:
                            - VLAN ID
                            - Not valid when type is INDEPENDENT, mainly
                              used to identify traffic from different ports
                              in container use case

                        type: int

                    type:
                        description: Type of port attachment.

                        type: str

                        choices:
                            - PARENT
                            - CHILD
                            - INDEPENDENT

            display_name:

                description:
                    - Segment Port display name.
                    - Either this or id must be specified. If both are
                      specified, id takes precedence.

                type: str

            description:

                description:
                    - Segment description.

                type: str

            extra_configs:

                description:
                    - Extra configs on segment port
                    - This property could be used for vendor specific
                      configuration in key value string pairs. Segment port
                      setting will override segment setting if the same key was
                      set on both segment and segment port.

                type: list
                element: dict
                suboptions:

                    config_pair:
                        description: Key value pair in string for the
                                     configuration

                        type: dict

                        required: true

                        suboptions:

                            key:
                                description: Key
                                type: str
                            value:
                                description: Value
                                type: str

            id:
                description: The id of the Policy Segment Port.

                type: str

            ignored_address_bindings:

                description:
                    - Address bindings to be ignored by IP Discovery module
                      IP Discovery module uses various mechanisms to discover
                      address bindings being used on each segment port. If a
                      user would like to ignore any specific discovered address
                      bindings or prevent the discovery of a particular set of
                      discovered bindings, then those address bindings can be
                      provided here. Currently IP range in CIDR format is not
                      supported.

                type: dict

                suboptions:

                ip_address:
                    description: IP Address for port binding.
                    type: str

            mac_address:

                description: Mac address for port binding.
                type: str

            vlan_id:
                description: VLAN ID for port binding.
                type: str

            state:

                choices:

                    - present
                    - absent

                description:

                    - State can be either 'present' or 'absent'. 'present' is
                      used to create or update resource. 'absent' is used to
                      delete resource
                    - Required if I(id != null)

            tags:

                description: Opaque identifiers meaningful to the API user.

                type: dict

                suboptions:

                    scope:
                        description: Tag scope.
                        type: str

                    tag:
                        description: Tag value.
                        type: str


    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    if state and str(state).lower() == "absent":
        ret["result"] = False
        ret["comment"] = (
            "Use absent method to delete segment resource. "
            "Only segment port is allowed to be deleted here."
        )
        return ret

    segment_response = __salt__["nsxt_policy_segment.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        display_name=display_name,
    )

    if "error" in segment_response:
        ret["result"] = False
        ret["comment"] = "Failed to get segment from NSX-T Manager : {}".format(
            segment_response["error"]
        )
        return ret

    result_count = len(segment_response["results"])

    if result_count > 1:
        ret["result"] = False
        ret["comment"] = "More than one segment exist with same display name : {}".format(
            display_name
        )
        return ret

    segment_dict = segment_response["results"][0] if result_count > 0 else None

    if __opts__["test"]:
        if result_count == 0:
            ret["result"] = None
            ret["comment"] = "Segment will be created in NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "Segment would be updated in NSX-T Manager"
        return ret

    if result_count == 0:
        # Create of segment will trigger in this flow
        log.info("Creating new segment in NSX-T with display_name %s", display_name)
        execution_logs_create = __salt__["nsxt_policy_segment.create_or_update"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            display_name=display_name,
            tags=tags,
            description=description,
            address_bindings=address_bindings,
            admin_state=admin_state,
            advanced_config=advanced_config,
            bridge_profiles=bridge_profiles,
            connectivity_path=connectivity_path,
            dhcp_config_path=dhcp_config_path,
            extra_configs=extra_configs,
            l2_extension=l2_extension,
            tier0_display_name=tier0_display_name,
            tier0_id=tier0_id,
            tier1_display_name=tier1_display_name,
            tier1_id=tier1_id,
            domain_name=domain_name,
            transport_zone_display_name=transport_zone_display_name,
            transport_zone_id=transport_zone_id,
            enforcementpoint_id=enforcementpoint_id,
            site_id=site_id,
            vlan_ids=vlan_ids,
            subnets=subnets,
            segment_ports=segment_ports,
        )

        log.info("Execution logs for segment create : {}".format(execution_logs_create))
        if "error" in execution_logs_create[len(execution_logs_create) - 1]:
            ret["result"] = False
            ret["comment"] = "Failed while doing create segment or its sub resource: {}".format(
                execution_logs_create
            )
            return ret

        segment_execution_log = next(
            (
                execution_log
                for execution_log in execution_logs_create
                if execution_log.get("resourceType") == "segments"
            ),
            None,
        )

        segment_id = segment_execution_log.get("results").get("id")

        segment_hierarchy = __salt__["nsxt_policy_segment.get_hierarchy"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            segment_id=segment_id,
        )
        if "error" in segment_hierarchy:
            ret["result"] = False
            ret["comment"] = "Failed while querying segment and its sub-resources: {}".format(
                segment_hierarchy["error"]
            )
            return ret
        ret["comment"] = "Created segment {display_name} successfully".format(
            display_name=display_name
        )
        ret["changes"]["new"] = segment_hierarchy
        return ret

    else:
        # Update of segment will be triggered
        segment_id = segment_dict.get("id")
        log.info("Updating existing segment with display_name %s", display_name)
        segment_hierarchy_response = __salt__["nsxt_policy_segment.get_hierarchy"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            segment_id=segment_id,
        )
        if "error" in segment_hierarchy_response:
            ret["result"] = False
            ret["comment"] = "Failed while querying segment and its sub-resources: {}".format(
                segment_hierarchy_response["error"]
            )
            return ret

        execution_logs_update = __salt__["nsxt_policy_segment.create_or_update"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            display_name=display_name,
            tags=tags,
            description=description,
            address_bindings=address_bindings,
            admin_state=admin_state,
            advanced_config=advanced_config,
            bridge_profiles=bridge_profiles,
            connectivity_path=connectivity_path,
            dhcp_config_path=dhcp_config_path,
            extra_configs=extra_configs,
            l2_extension=l2_extension,
            tier0_display_name=tier0_display_name,
            tier0_id=tier0_id,
            tier1_display_name=tier1_display_name,
            tier1_id=tier1_id,
            domain_name=domain_name,
            transport_zone_display_name=transport_zone_display_name,
            transport_zone_id=transport_zone_id,
            enforcementpoint_id=enforcementpoint_id,
            site_id=site_id,
            vlan_ids=vlan_ids,
            subnets=subnets,
            segment_ports=segment_ports,
        )

        log.info("Execution logs for updating segment : {}".format(execution_logs_update))
        if "error" in execution_logs_update[len(execution_logs_update) - 1]:
            ret["result"] = False
            ret["comment"] = "Failed while updating segment and sub-resources: {}".format(
                execution_logs_update
            )
            return ret

        segment_hierarchy_response_after_update = __salt__["nsxt_policy_segment.get_hierarchy"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            segment_id=segment_id,
        )
        if "error" in segment_hierarchy_response_after_update:
            ret["result"] = False
            ret["comment"] = "Failed while querying segment and its sub-resources: {}".format(
                segment_hierarchy_response_after_update["error"]
            )
            return ret

        ret["comment"] = "Updated segment {display_name} successfully".format(
            display_name=display_name
        )
        ret["changes"]["new"] = segment_hierarchy_response_after_update
        ret["changes"]["old"] = segment_hierarchy_response
        return ret


def absent(
    name,
    hostname,
    username,
    password,
    display_name,
    verify_ssl=None,
    cert=None,
    cert_common_name=None,
):
    """
    Deletes segment with the given display_name and all its sub-resources

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_policy_segment.absent hostname=nsxt-manager.local username=admin ...

        delete_segment:
          nsxt_policy_segment.absent:
            - name: <Name of the operation>
              hostname: <hostname>
              username: <username>
              password: <password>
              display_name: <display name of segment>
              cert: <certificate>
              verify_ssl: <False/True>

        name
            Name of the operation to perform

        hostname
            The host name of NSX-T manager

        username
            Username to connect to NSX-T manager

        password
            Password to connect to NSX-T manager

        display_name
            Display name of the segment to delete

        verify_ssl
            (Optional) Option to enable/disable SSL verification. Enabled by default.
            If set to False, the certificate validation is skipped.

        cert
            (Optional) Path to the SSL certificate file to connect to NSX-T manager.
            The certificate can be retrieved from browser.

        cert_common_name
            (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
             verification. If the client certificate common name and hostname do not match (in case of self-signed
             certificates), specify the certificate common name as part of this parameter. This value is then used to
             compare against certificate common name.

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    segment_response = __salt__["nsxt_policy_segment.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        display_name=display_name,
    )

    segment_dict, segment_id = None, None

    if "error" in segment_response:
        ret["result"] = False
        ret["comment"] = "Failed to get the segment response : {}".format(segment_response["error"])
        return ret

    segment_response_by_display_name = segment_response["results"]

    if len(segment_response_by_display_name) > 1:
        ret["result"] = False
        ret["comment"] = "More than one segment exist with same display name : {}".format(
            display_name
        )
        return ret

    segment_dict = (
        segment_response_by_display_name[0] if len(segment_response_by_display_name) > 0 else None
    )

    if segment_dict is not None:
        segment_id = segment_dict["id"]

    if segment_id is None:
        ret["comment"] = "No segment exist with display name %s" % display_name
        return ret

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if segment_dict is not None:
            ret["result"] = None
            ret["comment"] = "Segment will be deleted in NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "State absent will do nothing , since segment is not existing"
        return ret

    segment_hierarchy = __salt__["nsxt_policy_segment.get_hierarchy"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        segment_id=segment_id,
    )
    if "error" in segment_hierarchy:
        ret["result"] = False
        ret["comment"] = "Failure while querying segment and its sub-resources: {}".format(
            segment_hierarchy["error"]
        )
        return ret

    execution_logs_delete = __salt__["nsxt_policy_segment.delete"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        segment_id=segment_id,
    )
    log.info("Execution logs for deleting segment : {}".format(execution_logs_delete))
    if "error" in execution_logs_delete[len(execution_logs_delete) - 1]:
        ret["result"] = False
        ret["comment"] = "Failed to delete segment : {}".format(execution_logs_delete)
        return ret
    else:
        ret[
            "comment"
        ] = "Segment with display_name: {} and its sub-resources deleted successfully".format(
            display_name
        )
        ret["changes"]["old"] = segment_hierarchy
        ret["changes"]["new"] = {}
        return ret
